from typing import Any, Literal

from kirin import rewrite
from kirin.prelude import basic
from kirin.dialects import py, ilist


def test():
    rule = rewrite.Fixpoint(
        rewrite.Walk(
            rewrite.Chain(
                ilist.rewrite.HintLen(),
                rewrite.ConstantFold(),
                rewrite.DeadCodeElimination(),
            )
        )
    )

    @basic
    def len_func(xs: ilist.IList[int, Literal[3]]):
        return len(xs)

    @basic
    def len_func3(xs: ilist.IList[int, Any]):
        return len(xs)

    rule.rewrite(len_func.code)
    rule.rewrite(len_func3.code)

    stmt = len_func.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, py.Constant)
    assert stmt.value.unwrap() == 3
    assert len(len_func.callable_region.blocks[0].stmts) == 2

    stmt = len_func3.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, py.Len)
    assert len(len_func3.callable_region.blocks[0].stmts) == 2



from typing import Literal

from kirin import types
from kirin.prelude import basic, basic_no_opt
from kirin.analysis import TypeInference
from kirin.dialects import py, func, ilist


@basic_no_opt
def main(x):
    y: int = x
    return y


def test_ann_assign():
    stmt = main.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, py.assign.TypeAssert)

    typeinfer = TypeInference(basic_no_opt)
    _, ret = typeinfer.run(main, types.Int)
    assert ret.is_structurally_equal(types.Int)
    _, ret = typeinfer.run(main, types.Float)
    assert ret is ret.bottom()


def test_typeinfer_simplify_assert():
    @basic(typeinfer=True, fold=False)
    def simplify(x: int):
        y: int = x
        return y

    stmt = simplify.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, func.Return)


def test_list_assign():
    @basic_no_opt.add(ilist)
    def list_assign():
        xs: ilist.IList[float, Literal[3]] = ilist.IList([1, 2, 3], elem=types.Float)
        return xs

    stmt = list_assign.callable_region.blocks[0].stmts.at(3)
    assert isinstance(stmt, ilist.New)
    assert stmt.elem_type.is_structurally_equal(types.Float)
    assert stmt.result.type.is_structurally_equal(
        ilist.IListType[types.Float, types.Literal(3)]
    )

    stmt = list_assign.callable_region.blocks[0].stmts.at(4)
    assert isinstance(stmt, py.assign.TypeAssert)
    assert stmt.expected.is_structurally_equal(
        ilist.IListType[types.Float, types.Literal(3)]
    )



import random

from kirin.interp import Frame, MethodTable, impl

from . import stmts
from ._dialect import dialect


@dialect.register
class RandomMethodTable(MethodTable):

    @impl(stmts.Random)
    def random(self, interp, frame: Frame, stmt: stmts.Random):
        return (random.random(),)

    @impl(stmts.RandInt)
    def randint(self, interp, frame: Frame, stmt: stmts.RandInt):
        start = frame.get(stmt.start)
        stop = frame.get(stmt.stop)
        return (random.randint(start, stop),)

    @impl(stmts.Uniform)
    def uniform(self, interp, frame: Frame, stmt: stmts.Uniform):
        start = frame.get(stmt.start)
        stop = frame.get(stmt.stop)
        return (random.uniform(start, stop),)

    @impl(stmts.Seed)
    def seed(self, interp, frame: Frame, stmt: stmts.Seed):
        seed_value = frame.get(stmt.value)
        random.seed(seed_value)
        return tuple()



import pytest

from kirin.source import SourceInfo
from kirin.prelude import basic_no_opt


def get_line_of(target: str) -> int:
    for i, line in enumerate(open(__file__), 1):
        if target in line:
            return i


@pytest.mark.parametrize("similar", [True, False])
def test_stmt_source_info(similar: bool):
    @basic_no_opt
    def test(x: int):
        y = 2
        a = 4**2
        return y + 2 + a

    if similar:
        test = test.similar()

    stmts = test.callable_region.blocks[0].stmts

    def get_line_from_source_info(source: SourceInfo) -> int:
        return source.lineno + source.lineno_begin

    for stmt in stmts:
        assert stmt.source.file == __file__

    assert get_line_from_source_info(stmts.at(0).source) == get_line_of("y = 2")
    assert get_line_from_source_info(stmts.at(2).source) == get_line_of("a = 4**2")
    assert get_line_from_source_info(stmts.at(4).source) == get_line_of(
        "return y + 2 + a"
    )



from dataclasses import dataclass

from .abc import Attribute
from .types import (
    Union,
    Generic,
    Literal,
    PyClass,
    TypeVar,
    FunctionType,
    TypeAttribute,
    TypeofMethodType,
)

@dataclass
class _TypeAttribute(Attribute):
    def is_subseteq_Union(self, other: Union) -> bool: ...
    def is_subseteq_Literal(self, other: Literal) -> bool: ...
    def is_subseteq_TypeVar(self, other: TypeVar) -> bool: ...
    def is_subseteq_PyClass(self, other: PyClass) -> bool: ...
    def is_subseteq_Generic(self, other: Generic) -> bool: ...
    def is_subseteq_fallback(self, other: TypeAttribute) -> bool: ...
    def is_subseteq_TypeofMethodType(self, other: TypeofMethodType) -> bool: ...
    def is_subseteq_FunctionType(self, other: FunctionType) -> bool: ...



from typing import Callable
from dataclasses import field, dataclass

from kirin import ir
from kirin.passes import Pass
from kirin.rewrite import Walk, Inline, Fixpoint, CFGCompactify, DeadCodeElimination
from kirin.rewrite.abc import RewriteResult


def aggresive(x: ir.IRNode) -> bool:
    return True


@dataclass
class InlinePass(Pass):
    heuristic: Callable[[ir.IRNode], bool] = field(default=aggresive)

    def unsafe_run(self, mt: ir.Method) -> RewriteResult:

        result = Walk(Inline(heuristic=self.heuristic)).rewrite(mt.code)
        result = Walk(CFGCompactify()).rewrite(mt.code).join(result)

        # dce
        dce = DeadCodeElimination()
        return Fixpoint(Walk(dce)).rewrite(mt.code).join(result)



from typing import Any

import numpy as np
import pytest

from kirin.prelude import basic
from kirin.dialects import vmath
from kirin.dialects.ilist.runtime import IList


@basic.union([vmath])
def add_kernel(x, y):
    return x + y


@basic.union([vmath])(typeinfer=True)
def add_scalar_rhs_typed(x: IList[float, Any], y: float):
    return x + y


@basic.union([vmath])(aggressive=True, typeinfer=True)
def add_scalar_lhs():
    return add_kernel(x=3.0, y=[3.0, 4, 5])


@pytest.mark.xfail()
def test_add_scalar_lhs():
    # out = add_scalar_lhs()
    add_scalar_lhs.print()
    res = add_scalar_lhs()
    assert isinstance(res, IList)
    assert res.type.vars[0].typ is float
    assert np.allclose(np.asarray(res), np.array([6, 7, 8]))


@pytest.mark.xfail()
def test_typed_kernel_add():
    add_scalar_rhs_typed.print()
    res = add_scalar_rhs_typed(IList([0.0, 1.0, 2.0]), 3.1)
    assert np.allclose(np.asarray(res), np.asarray([3.1, 4.1, 5.1]))


@basic.union([vmath])(typeinfer=True)
def add_two_lists():
    return add_kernel(x=[0, 1, 2], y=[3, 4, 5])


def test_add_lists():
    res = add_two_lists()
    assert np.allclose(np.asarray(res), np.array([0, 1, 2, 3, 4, 5]))


@basic.union([vmath])(typeinfer=True)
def sub_scalar_rhs_typed(x: IList[float, Any], y: float):
    return x - y


@pytest.mark.xfail()
def test_sub_scalar_typed():
    res = sub_scalar_rhs_typed(IList([0, 1, 2]), 3.1)
    assert np.allclose(np.asarray(res), np.asarray([-3.1, -2.1, -1.1]))


@basic.union([vmath])(typeinfer=True)
def mult_scalar_lhs_typed(x: float, y: IList[float, Any]):
    return x * y


@basic.union([vmath])(typeinfer=True)
def mult_kernel(x, y):
    return x * y


@basic.union([vmath])(typeinfer=True, aggressive=True)
def mult_scalar_lhs():
    return mult_kernel(x=3.0, y=[3.0, 4.0, 5.0])


@pytest.mark.xfail()
def test_mult_scalar_typed():
    res = mult_scalar_lhs_typed(3, IList([0, 1, 2]))
    assert np.allclose(np.asarray(res), np.asarray([0, 3, 6]))


@pytest.mark.xfail()
def test_mult_scalar_lhs():
    res = mult_scalar_lhs()
    assert isinstance(res, IList)
    assert res.type.vars[0].typ is float
    assert np.allclose(np.asarray(res), np.array([9, 12, 15]))


@basic.union([vmath])(typeinfer=True)
def div_scalar_lhs_typed(x: float, y: IList[float, Any]):
    return x / y


def test_div_scalar_typed():
    res = div_scalar_lhs_typed(3, IList([1, 1.5, 2]))
    assert np.allclose(np.asarray(res), np.asarray([3, 2, 1.5]))



# type: ignore

# [section]
from enum import Enum
from typing import ClassVar
from dataclasses import dataclass

from qulacs import QuantumState


# this could be your own class implementing the runtime in whatever way you want
@dataclass
class Qubit:
    count: ClassVar[int] = 0  # class variable to count qubits
    id: int

    def __init__(self):
        self.id = Qubit.count
        Qubit.count += 1


# some your own classes
class Basis(Enum):
    X = "X"
    Y = "Y"
    Z = "Z"


# [section]
from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.prelude import basic

# our language definitions and compiler begins
dialect = ir.Dialect("quantum")
QubitType = types.PyClass(Qubit)
StateType = types.PyClass(QuantumState)


@statement(dialect=dialect)
class NewQubit(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})
    state: ir.SSAValue = info.argument(
        StateType
    )  # we can use Python objects as arguments
    qubit: ir.ResultValue = info.result(QubitType)


@statement(dialect=dialect)
class X(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})
    state: ir.SSAValue = info.argument(StateType)
    qubit: ir.SSAValue = info.argument(QubitType)


@statement(dialect=dialect)
class H(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})
    state: ir.SSAValue = info.argument(StateType)
    qubit: ir.SSAValue = info.argument(QubitType)


@statement(dialect=dialect)
class CX(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})
    state: ir.SSAValue = info.argument(StateType)
    control: ir.SSAValue = info.argument(QubitType)
    target: ir.SSAValue = info.argument(QubitType)


@statement(dialect=dialect)
class CZ(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})
    state: ir.SSAValue = info.argument(StateType)
    control: ir.SSAValue = info.argument(QubitType)
    target: ir.SSAValue = info.argument(QubitType)


@statement(dialect=dialect)
class Measure(ir.Statement):
    traits = frozenset({lowering.FromPythonCall()})
    basis: Basis = (
        info.attribute()
    )  # we can use Python objects as attributes (compile-time values)!
    state: ir.SSAValue = info.argument(StateType)
    qubit: ir.SSAValue = info.argument(QubitType)
    result: ir.ResultValue = info.result(types.Int)


# now we have the miminim set of statements to represent a quantum circuit
# the following defines a group of "dialects" so we can use it as a decorator
@ir.dialect_group(basic.add(dialect))
def quantum(self):  # group self
    def run_default_pass(method, option_a=True):
        # default pass to run right after calling the decorator
        # a.k.a the default JIT compilation part of the compiler
        pass

    return run_default_pass


# Ok let's try it out
@quantum
def main(state: QuantumState):
    a = NewQubit(state)
    b = NewQubit(state)
    H(state, a)
    CX(state, control=a, target=b)
    return Measure(state, basis=Basis.Z, qubit=b)


# well Linter is mad at us

# [section]
# fortunately, Kirin provides a way to give hints to a standard Python linter
# now let's make some lowering wrappers to make Python type hinting happy


@lowering.wraps(NewQubit)
def new_qubit(state: QuantumState) -> Qubit: ...


@lowering.wraps(X)
def x(state: QuantumState, qubit: Qubit) -> None: ...


@lowering.wraps(H)
def h(state: QuantumState, qubit: Qubit) -> None: ...


@lowering.wraps(CX)
def cx(state: QuantumState, control: Qubit, target: Qubit) -> None: ...


@lowering.wraps(CZ)
def cz(state: QuantumState, control: Qubit, target: Qubit) -> None: ...


@lowering.wraps(Measure)
def measure(state: QuantumState, basis: Basis, qubit: Qubit) -> None: ...


# this is a lot nicer now!
@quantum
def main(state: QuantumState):
    a = new_qubit(state)
    b = new_qubit(state)
    h(state, a)
    h(state, b)
    cx(state, control=a, target=b)
    if measure(state, basis=Basis.Z, qubit=b):
        x(state, a)  # we can use the result of Measure to conditionally apply X gate
    return


main.print()

# Ok but this doesn't work yet, I cannot run it
# main()

# [section]
# we need to implement the runtime for the quantum circuit
# let's just import qulacs a quantum circuit simulator

from qulacs import QuantumState, gate

from kirin import interp


@dialect.register
class MethodTable(interp.MethodTable):
    @interp.impl(NewQubit)
    def impl_new_qubit(
        self, interp: interp.Interpreter, frame: interp.Frame, stmt: NewQubit
    ) -> tuple[Qubit]:
        return (Qubit(),)

    @interp.impl(X)
    def impl_x(self, interp: interp.Interpreter, frame: interp.Frame, stmt: X) -> None:
        state = frame.get_casted(
            stmt.state, QuantumState
        )  # assume state is QuantumState at runtime
        qubit = frame.get_casted(
            stmt.qubit, Qubit
        )  # we assume qubits are Qubit at runtime
        gate.X(qubit.id).update_quantum_state(state)

    @interp.impl(H)
    def impl_h(self, interp: interp.Interpreter, frame: interp.Frame, stmt: H) -> None:
        state = frame.get_casted(stmt.state, QuantumState)
        qubit = frame.get_casted(stmt.qubit, Qubit)
        gate.H(qubit.id).update_quantum_state(state)

    @interp.impl(CX)
    def impl_cx(
        self, interp: interp.Interpreter, frame: interp.Frame, stmt: CX
    ) -> None:
        state = frame.get_casted(stmt.state, QuantumState)
        control = frame.get_casted(stmt.control, Qubit)
        target = frame.get_casted(stmt.target, Qubit)
        print(f"Applying CNOT gate with control {control.id} and target {target.id}")
        gate.CNOT(control.id, target.id).update_quantum_state(state)

    @interp.impl(CZ)
    def impl_cz(
        self, interp: interp.Interpreter, frame: interp.Frame, stmt: CZ
    ) -> None:
        state = frame.get_casted(stmt.state, QuantumState)
        control = frame.get_casted(stmt.control, Qubit)
        target = frame.get_casted(stmt.target, Qubit)
        print(f"Applying CZ gate with control {control.id} and target {target.id}")
        gate.CZ(control.id, target.id).update_quantum_state(state)

    @interp.impl(Measure)
    def impl_measure(
        self, interp: interp.Interpreter, frame: interp.Frame, stmt: Measure
    ) -> tuple[int]:
        state = frame.get_casted(stmt.state, QuantumState)
        qubit = frame.get_casted(stmt.qubit, Qubit)
        basis = stmt.basis.value  # get the basis as a string
        result = gate.Measurement(qubit.id, qubit.id).update_quantum_state(state)
        return (
            state.get_classical_value(qubit.id),
        )  # return the measurement result as an int


state = QuantumState(2)  # 2 qubits
state.set_zero_state()
main(state)
print(state.get_vector())

# [section]
# ok now we can run it, how about rewriting the program?

from kirin.rewrite.abc import RewriteRule, RewriteResult


class CX2CZ(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, CX):
            return RewriteResult()

        H(node.state, node.target).insert_before(node)
        node.replace_by(
            cz_node := CZ(state=node.state, control=node.control, target=node.target)
        )
        H(node.state, node.target).insert_after(cz_node)
        return RewriteResult(has_done_something=True)


from kirin.rewrite import Walk

Walk(CX2CZ()).rewrite(main.code)
main.print()



from __future__ import annotations

import sys
from typing import IO, Generic, TypeVar, cast
from contextlib import contextmanager
from dataclasses import dataclass

from kirin import ir, interp

from .abc import EmitABC, EmitFrame

IO_t = TypeVar("IO_t", bound=IO)


@dataclass
class JuliaFrame(EmitFrame[str], Generic[IO_t]):
    io: IO_t = cast(IO_t, sys.stdout)

    def write(self, value):
        self.io.write(value)

    def write_line(self, value):
        self.write("    " * self._indent + value + "\n")

    @contextmanager
    def indent(self):
        self._indent += 1
        yield
        self._indent -= 1


@dataclass
class Julia(EmitABC[JuliaFrame, str], Generic[IO_t]):
    """Julia code generator for the IR.

    This class generates Julia code from the IR.
    It is used to generate Julia code for the IR.
    """

    keys = ("emit.julia",)
    void = ""

    # some states
    io: IO_t

    def initialize(self):
        super().initialize()
        return self

    def initialize_frame(
        self, node: ir.Statement, *, has_parent_access: bool = False
    ) -> JuliaFrame:
        return JuliaFrame(node, self.io, has_parent_access=has_parent_access)

    def frame_call(
        self, frame: JuliaFrame, node: ir.Statement, *args: str, **kwargs: str
    ) -> str:
        return f"{args[0]}({', '.join(args[1:])})"

    def get_attribute(self, frame: JuliaFrame, node: ir.Attribute) -> str:
        method = self.registry.get(interp.Signature(type(node)))
        if method is None:
            raise ValueError(f"Method not found for node: {node}")
        return method(self, frame, node)

    def reset(self):
        self.io.truncate(0)
        self.io.seek(0)



from kirin import ir, types
from kirin.analysis import const
from kirin.dialects import py
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.ilist.stmts import IListType

from .._dialect import dialect


@dialect.post_inference
class HintLen(RewriteRule):

    def _get_collection_len(self, collection: ir.SSAValue):
        coll_type = collection.type

        if not isinstance(coll_type, types.Generic):
            return None

        if (
            coll_type.is_subseteq(IListType)
            and isinstance(coll_type.vars[1], types.Literal)
            and isinstance(coll_type.vars[1].data, int)
        ):
            return coll_type.vars[1].data
        else:
            return None

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, py.Len):
            return RewriteResult()

        if (coll_len := self._get_collection_len(node.value)) is None:
            return RewriteResult()

        existing_hint = node.result.hints.get("const")
        new_hint = const.Value(coll_len)

        if existing_hint is not None and new_hint.is_structurally_equal(existing_hint):
            return RewriteResult()

        node.result.hints["const"] = new_hint
        return RewriteResult(has_done_something=True)



from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.prelude import basic_no_opt
from kirin.analysis import const

dialect = ir.Dialect("mwe")


@statement(dialect=dialect)
class SideEffect(ir.Statement):
    name = "side_effect"
    traits = frozenset({lowering.FromPythonCall()})
    value: ir.SSAValue = info.argument(types.Int)


@basic_no_opt.add(dialect)
def recursion(kernel, n: int, pos: int):
    if pos == n:
        return

    kernel(pos)
    recursion(kernel, n, pos + 1)


@basic_no_opt.add(dialect)
def side_effect(pos: int):
    SideEffect(pos)  # type: ignore


def test_non_pure_const():
    constprop = const.Propagate(basic_no_opt)
    frame, ret = constprop.run(
        recursion,
        const.Value(side_effect),
        const.Result.top(),
        const.Result.top(),
    )
    # recursion.print(analysis=frame.entries)
    ret = frame.entries[recursion.callable_region.blocks[2].stmts.at(3).results[0]]
    assert isinstance(ret, const.Value)
    assert frame.frame_is_not_pure



from kirin import ir, types
from kirin.prelude import basic


def test_method_type_hint():
    @basic
    def main() -> ir.Method[[int, int], float]:

        def test(x: int, y: int) -> float:
            return x * y * 3.0

        return test

    assert main.return_type == types.MethodType[[types.Int, types.Int], types.Float]



import pytest

from kirin.prelude import basic_no_opt
from kirin.rewrite import Walk, Chain, Fixpoint, WrapConst
from kirin.analysis import const
from kirin.rewrite.dce import DeadCodeElimination
from kirin.rewrite.getitem import InlineGetItem
from kirin.dialects.py.indexing import GetItem


def apply_getitem_optimization(func):
    constprop = const.Propagate(func.dialects)
    frame, _ = constprop.run(func)
    Fixpoint(Walk(WrapConst(frame))).rewrite(func.code)
    inline_getitem = InlineGetItem()
    Fixpoint(Walk(Chain([inline_getitem, DeadCodeElimination()]))).rewrite(func.code)


@pytest.mark.parametrize("index", [0, -1, 1])
def test_getitem_index(index):

    @basic_no_opt
    def func(x: int):
        ylist = (x, x, 1, x)
        return ylist[index]

    before = func(1)
    apply_getitem_optimization(func)
    after = func(1)

    assert before == after
    assert len(func.callable_region.blocks[0].stmts) == 1


@pytest.mark.parametrize(
    "sl",
    [
        slice(0, 2, 1),
        slice(None, None, None),
        slice(None, -1, None),
        slice(-1, None, None),
        slice(None, None, -1),
        slice(1, 4, 2),
    ],
)
def test_getitem_slice(sl):

    @basic_no_opt
    def func():
        ylist = (0, 1, 2, 3, 4)
        return ylist[sl]

    stmt_types = [type(stmt) for stmt in func.callable_region.blocks[0].stmts]
    assert GetItem in stmt_types

    before = func()
    apply_getitem_optimization(func)
    after = func()

    assert before == after
    stmt_types = [type(stmt) for stmt in func.callable_region.blocks[0].stmts]
    assert GetItem not in stmt_types


@pytest.mark.parametrize(
    "start, stop, step",
    [
        (0, 2, 1),
        (None, None, None),
        (None, -1, None),
        (-1, None, None),
        (None, None, -1),
        (1, 4, 2),
    ],
)
def test_getitem_slice_with_literal_indices(start, stop, step):

    @basic_no_opt
    def func():
        ylist = (0, 1, 2, 3, 4)
        return ylist[start:stop:step]

    stmt_types = [type(stmt) for stmt in func.callable_region.blocks[0].stmts]
    assert GetItem in stmt_types

    before = func()

    apply_getitem_optimization(func)

    stmt_types = [type(stmt) for stmt in func.callable_region.blocks[0].stmts]
    assert GetItem not in stmt_types
    after = func()

    assert before == after


def test_getitem_slice_does_nothing():
    """Test that inline_getiem returns `has_done_something=False` when the getitem
    statement has no uses"""

    @basic_no_opt
    def func():
        ylist = (0, 1, 2, 3, 4)
        ylist[0]  # no use
        return 1

    constprop = const.Propagate(func.dialects)
    frame, _ = constprop.run(func)
    Fixpoint(Walk(WrapConst(frame))).rewrite(func.code)

    inline_getitem = InlineGetItem()

    assert not Walk(inline_getitem).rewrite(func.code).has_done_something



from kirin import ir
from kirin.passes import TypeInfer
from kirin.dialects import py
from kirin.rewrite.abc import RewriteRule, RewriteResult

from ..stmts import Lambda, Function, GetField
from .._dialect import dialect


@dialect.canonicalize
class LambdaLifting(RewriteRule):
    """Lifts func.Lambda methods embedded in py.Constant into func.Function.
    - Trigger on py.Constant
    """

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        from kirin.dialects import py

        if not isinstance(node, py.Constant):
            return RewriteResult(has_done_something=False)
        method = self._get_method_from_constant(node)
        if method is None:
            return RewriteResult(has_done_something=False)
        if not isinstance(method.code, Lambda):
            return RewriteResult(has_done_something=False)
        self._promote_lambda(method)

        rewrite_result = TypeInfer(dialects=method.dialects).unsafe_run(method)
        return RewriteResult(has_done_something=True).join(rewrite_result)

    def _get_method_from_constant(self, const_stmt: py.Constant) -> ir.Method | None:
        pyattr_data = const_stmt.value
        if isinstance(pyattr_data, ir.PyAttr) and isinstance(
            pyattr_data.data, ir.Method
        ):
            return pyattr_data.data
        return None

    def _get_field_index(self, getfield_stmt: GetField) -> int | None:
        fld = getfield_stmt.attributes.get("field")
        if fld:
            return getfield_stmt.field
        else:
            return None

    def _promote_lambda(self, method: ir.Method) -> None:
        new_method = method.similar()
        assert isinstance(
            new_method.code, Lambda
        ), "expected method.code to be func.Lambda before promotion"

        captured_fields = method.fields
        if captured_fields:
            for stmt in new_method.code.body.blocks[0].stmts:
                if not isinstance(stmt, GetField):
                    continue
                idx = self._get_field_index(stmt)
                if idx is None:
                    continue
                captured = new_method.fields[idx]
                from kirin.dialects import py

                const_stmt = py.Constant(captured)
                const_stmt.insert_before(stmt)
                if stmt.results and const_stmt.results:
                    stmt.results[0].replace_by(const_stmt.results[0])
                stmt.delete()
                new_method.code

        fn = Function(
            sym_name=new_method.code.sym_name,
            slots=new_method.code.slots,
            signature=new_method.code.signature,
            body=new_method.code.body,
        )
        method.code = fn



from typing import cast

from kirin import ir, types
from kirin.ir import Block, Region
from kirin.decl import info, statement
from kirin.print.printer import Printer

from ._dialect import dialect


@statement(dialect=dialect, init=False)
class IfElse(ir.Statement):
    """Python-like if-else statement.

    This statement has a condition, then body, and else body.

    Then body either terminates with a yield statement or `scf.return`.
    """

    name = "if"
    traits = frozenset({ir.MaybePure(), ir.HasCFG(), ir.SSACFG()})
    purity: bool = info.attribute(default=False)
    cond: ir.SSAValue = info.argument(types.Any)
    # NOTE: we don't enforce the type here
    # because anything implements __bool__ in Python
    # can be used as a condition
    then_body: ir.Region = info.region(multi=False)
    else_body: ir.Region = info.region(multi=False, default_factory=ir.Region)

    def __init__(
        self,
        cond: ir.SSAValue,
        then_body: ir.Region | ir.Block,
        else_body: ir.Region | ir.Block | None = None,
    ):
        if then_body.IS_REGION:
            then_body_region = cast(Region, then_body)
            if then_body_region.blocks:
                then_body_block = then_body_region.blocks[-1]
            else:
                then_body_block = None
        else:  # then_body.IS_BLOCK:
            then_body_block = cast(Block, then_body)
            then_body_region = Region(then_body_block)

        if else_body is None:
            else_body_region = ir.Region()
            else_body_block = None
        elif else_body.IS_REGION:
            else_body_region = cast(Region, else_body)
            if not else_body_region.blocks:  # empty region
                else_body_block = None
            elif len(else_body_region.blocks) == 0:
                else_body_block = None
            else:
                else_body_block = else_body_region.blocks[0]
        else:  # else_body.IS_BLOCK:
            else_body_region = ir.Region(cast(Block, else_body))
            else_body_block = else_body

        # if either then or else body has yield, we generate results
        # we assume if both have yields, they have the same number of results
        results = ()
        if then_body_block is not None:
            then_yield = then_body_block.last_stmt
            else_body_block = cast(Block, else_body_block)
            else_yield = (
                else_body_block.last_stmt if else_body_block is not None else None
            )
            if then_yield is not None and isinstance(then_yield, Yield):
                results = then_yield.values
            elif else_yield is not None and isinstance(else_yield, Yield):
                results = else_yield.values

        result_types = tuple(value.type for value in results)
        super().__init__(
            args=(cond,),
            regions=(then_body_region, else_body_region),
            result_types=result_types,
            args_slice={"cond": 0},
            attributes={"purity": ir.PyAttr(False)},
        )

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        printer.print(self.cond)
        printer.plain_print(" ")
        printer.print(self.then_body)
        if self.else_body.blocks and not (
            len(self.else_body.blocks[0].stmts) == 1
            and isinstance(else_term := self.else_body.blocks[0].last_stmt, Yield)
            and not else_term.values  # empty yield
        ):
            printer.plain_print(" else ", style="keyword")
            printer.print(self.else_body)

        with printer.rich(style="comment"):
            printer.plain_print(f" -> purity={self.purity}")

    def verify(self) -> None:
        from kirin.dialects.func import Return

        if len(self.then_body.blocks) != 1:
            raise ir.ValidationError(self, "then region must have a single block")

        if len(self.else_body.blocks) != 1:
            raise ir.ValidationError(self, "else region must have a single block")

        then_block = self.then_body.blocks[0]
        else_block = self.else_body.blocks[0]
        if len(then_block.args) != 1:
            raise ir.ValidationError(
                self, "then block must have a single argument for condition"
            )

        if len(else_block.args) != 1:
            raise ir.ValidationError(
                self, "else block must have a single argument for condition"
            )

        then_stmt = then_block.last_stmt
        else_stmt = else_block.last_stmt
        if then_stmt is None or not isinstance(then_stmt, (Yield, Return)):
            raise ir.ValidationError(
                self, "then block must terminate with a yield or return"
            )

        if else_stmt is None or not isinstance(else_stmt, (Yield, Return)):
            raise ir.ValidationError(
                self, "else block must terminate with a yield or return"
            )


@statement(dialect=dialect, init=False)
class For(ir.Statement):
    name = "for"
    traits = frozenset({ir.MaybePure(), ir.HasCFG(), ir.SSACFG()})
    purity: bool = info.attribute(default=False)
    iterable: ir.SSAValue = info.argument(types.Any)
    body: ir.Region = info.region(multi=False)
    initializers: tuple[ir.SSAValue, ...] = info.argument(types.Any)

    def __init__(
        self,
        iterable: ir.SSAValue,
        body: ir.Region,
        *initializers: ir.SSAValue,
    ):
        stmt = body.blocks[0].last_stmt
        if isinstance(stmt, Yield):
            result_types = tuple(value.type for value in stmt.values)
        else:
            result_types = ()
        super().__init__(
            args=(iterable, *initializers),
            regions=(body,),
            result_types=result_types,
            args_slice={"iterable": 0, "initializers": slice(1, None)},
            attributes={"purity": ir.PyAttr(False)},
        )

    def verify(self) -> None:
        from kirin.dialects.func import Return

        if len(self.body.blocks) != 1:
            raise ir.ValidationError(self, "for loop body must have a single block")

        if len(self.body.blocks[0].args) != len(self.initializers) + 1:
            raise ir.ValidationError(
                self,
                "for loop body must have arguments for all initializers and the loop variable",
            )

        stmt = self.body.blocks[0].last_stmt
        if stmt is None or not isinstance(stmt, (Yield, Return)):
            raise ir.ValidationError(
                self, "for loop body must terminate with a yield or return"
            )

        if isinstance(stmt, Return):
            return

        if len(stmt.values) != len(self.initializers):
            raise ir.ValidationError(
                self,
                "for loop body must have the same number of results as initializers",
            )
        if len(self.results) != len(stmt.values):
            raise ir.ValidationError(
                self,
                "for loop must have the same number of results as the yield in the body",
            )

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        block = self.body.blocks[0]
        printer.print(block.args[0])
        printer.plain_print(" in ", style="keyword")
        printer.print(self.iterable)
        if self.results:
            with printer.rich(style="comment"):
                printer.plain_print(" -> ")
                printer.print_seq(
                    tuple(result.type for result in self.results),
                    delim=", ",
                    style="comment",
                )

        with printer.indent():
            if self.initializers:
                printer.print_newline()
                printer.plain_print("iter_args(")
                for idx, (arg, val) in enumerate(
                    zip(block.args[1:], self.initializers)
                ):
                    printer.print(arg)
                    printer.plain_print(" = ")
                    printer.print(val)
                    if idx < len(self.initializers) - 1:
                        printer.plain_print(", ")
                printer.plain_print(")")

            printer.plain_print(" {")
            if printer.analysis is not None:
                with printer.rich(style="warning"):
                    for arg in block.args:
                        printer.print_newline()
                        printer.print_analysis(
                            arg, prefix=f"{printer.state.ssa_id[arg]} --> "
                        )
            with printer.align(printer.result_width(block.stmts)):
                for stmt in block.stmts:
                    printer.print_newline()
                    printer.print_stmt(stmt)
        printer.print_newline()
        printer.plain_print("}")
        with printer.rich(style="comment"):
            printer.plain_print(f" -> purity={self.purity}")


@statement(dialect=dialect)
class Yield(ir.Statement):
    name = "yield"
    traits = frozenset({ir.IsTerminator()})
    values: tuple[ir.SSAValue, ...] = info.argument(types.Any)

    def __init__(self, *values: ir.SSAValue):
        super().__init__(args=values, args_slice={"values": slice(None)})

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.print_seq(self.values, prefix=" ", delim=", ")



from kirin import ir, types
from kirin.dialects import func


def test_is_structurally_equal_ignoring_hint():
    block = ir.Block()
    block.args.append_from(types.MethodType, "self")
    source_func = func.Function(
        sym_name="main",
        signature=func.Signature(
            inputs=(),
            output=types.NoneType,
        ),
        body=ir.Region(block),
    )

    block = ir.Block()
    block.args.append_from(types.MethodType, "self")
    expected_func = func.Function(
        sym_name="main",
        signature=func.Signature(
            inputs=(),
            output=types.NoneType,
        ),
        body=ir.Region(block),
    )

    assert expected_func.is_structurally_equal(source_func)



from pathlib import Path

from kirin import emit
from kirin.prelude import structural
from kirin.dialects import debug


@structural.add(debug)
def some_arith(x: int, y: float):
    return x + y


@structural.add(debug)
def julia_like(x: int, y: int):
    for i in range(x):
        for j in range(y):
            if i == 0:
                debug.info("Hello")
            else:
                debug.info("World")
    return some_arith(x + y, 4.0)


def test_julia_like(tmp_path):
    file = tmp_path / "julia_like.jl"
    with open(file, "w") as io:
        julia_emit = emit.Julia(structural.add(debug), io=io)
        julia_emit.run(julia_like)

    with open(file, "r") as io:
        generated = io.read()

    with open(Path(__file__).parent / "julia_like.jl", "r") as io:
        target = io.read()

    assert generated.strip() == target.strip()



from dataclasses import dataclass

from kirin import ir
from kirin.analysis import const
from kirin.dialects import py
from kirin.rewrite.abc import RewriteRule, RewriteResult


@dataclass
class InlineGetItem(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, py.indexing.GetItem):
            return RewriteResult()

        if not isinstance(node.obj.owner, py.tuple.New):
            return RewriteResult()

        if not isinstance(index_value := node.index.hints.get("const"), const.Value):
            return RewriteResult()

        if not node.result.uses:
            return RewriteResult()

        stmt = node.obj.owner
        index = index_value.data
        if isinstance(index, int) and (
            0 <= index < len(stmt.args) or -len(stmt.args) <= index < 0
        ):
            node.result.replace_by(stmt.args[index])
            return RewriteResult(has_done_something=True)
        elif isinstance(index, slice):
            new_tuple = py.tuple.New(tuple(stmt.args[index]))
            node.replace_by(new_tuple)
            return RewriteResult(has_done_something=True)
        else:
            return RewriteResult()



from __future__ import annotations

from queue import SimpleQueue
from typing import Generic, TypeVar, Iterable

ElemType = TypeVar("ElemType")


class WorkList(SimpleQueue, Generic[ElemType]):
    """The worklist data structure.

    The worklist is a stack that allows for O(1) removal of elements from the stack.
    """

    def __len__(self) -> int:
        return self.qsize()

    def __bool__(self) -> bool:
        return not self.empty()

    def is_empty(self) -> bool:
        return self.empty()

    def append(self, item: ElemType) -> None:
        self.put_nowait(item)

    def extend(self, items: Iterable[ElemType]) -> None:
        for item in items:
            self.put_nowait(item)

    def pop(self) -> ElemType | None:
        if self.empty():
            return None
        return self.get_nowait()


# Remove one function call from critical speed bottleneck
WorkList.is_empty = WorkList.empty
WorkList.append = WorkList.put_nowait



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, ClassVar, cast
from dataclasses import field, dataclass

from typing_extensions import Self

from kirin.print import Printer, Printable
from kirin.ir.attrs.abc import Attribute
from kirin.ir.attrs.types import AnyType, TypeAttribute

if TYPE_CHECKING:
    from kirin.ir.use import Use
    from kirin.ir.nodes.stmt import Statement
    from kirin.ir.nodes.block import Block
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.deserializer import Deserializer
    from kirin.serialization.core.serializationunit import SerializationUnit


@dataclass
class SSAValue(ABC, Printable):
    """Base class for all SSA values in the IR."""

    IS_SSA_VALUE: ClassVar[bool] = True

    type: TypeAttribute = field(default_factory=AnyType, init=False, repr=True)
    """The type of this SSA value."""
    hints: dict[str, Attribute] = field(default_factory=dict, init=False, repr=False)
    """Hints for this SSA value."""
    uses: set[Use] = field(init=False, default_factory=set, repr=False)
    """The uses of this SSA value."""
    _name: str | None = field(init=False, default=None, repr=True)
    """The name of this SSA value."""
    name_pattern: ClassVar[re.Pattern[str]] = re.compile(r"([A-Za-z_$.-][\w$.-]*)")
    """The pattern that the name of this SSA value must match."""

    @property
    @abstractmethod
    def owner(self) -> Statement | Block:
        """The object that owns this SSA value."""
        ...

    @property
    def name(self) -> str | None:
        """The name of this SSA value."""
        return self._name

    @name.setter
    def name(self, name: str | None) -> None:
        if name and not self.name_pattern.fullmatch(name):
            raise ValueError(f"Invalid name: {name}")
        self._name = name

    def __repr__(self) -> str:
        if self.name:
            return f"{type(self).__name__}({self.name})"
        return f"{type(self).__name__}({id(self)})"

    def __hash__(self) -> int:
        return id(self)

    def add_use(self, use: Use) -> Self:
        """Add a use to this SSA value."""
        self.uses.add(use)
        return self

    def remove_use(self, use: Use) -> Self:
        """Remove a use from this SSA value."""
        # print(use)
        # assert use in self.uses, "Use not found"
        if use in self.uses:
            self.uses.remove(use)
        return self

    def replace_by(self, other: SSAValue) -> None:
        """Replace this SSA value with another SSA value. Update all uses."""
        for use in self.uses.copy():
            use.stmt.args[use.index] = other

        if other.name is None and self.name is not None:
            other.name = self.name

        assert len(self.uses) == 0, "Uses not empty"

    # TODO: also delete BlockArgument from arglist
    def delete(self, safe: bool = True) -> None:
        """Delete this SSA value. If `safe` is `True`, raise an error if there are uses."""
        if safe and len(self.uses) > 0:
            raise ValueError("Cannot delete SSA value with uses")
        self.replace_by(DeletedSSAValue(self))

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print(printer.state.ssa_id[self])


@dataclass
class ResultValue(SSAValue):
    """SSAValue that is a result of a [`Statement`][kirin.ir.nodes.stmt.Statement]."""

    stmt: Statement = field(init=False)
    """The statement that this value is a result of."""
    index: int = field(init=False)
    """The index of this value in the statement's result list."""

    # NOTE: we will assign AnyType unless specified.
    #       when SSAValue is a ResultValue, the type is inferred
    #       later in the compilation process.
    def __init__(
        self, stmt: Statement, index: int, type: TypeAttribute | None = None
    ) -> None:
        super().__init__()
        self.type = type or AnyType()
        self.stmt = stmt
        self.index = index

    @property
    def owner(self) -> Statement:
        return self.stmt

    def __hash__(self) -> int:
        return id(self)

    def __repr__(self) -> str:
        if self.type is self.type.top():
            type_str = ""
        else:
            type_str = f"[{self.type}]"

        if self.name:
            return (
                f"<{type(self).__name__}{type_str} {self.name}, uses: {len(self.uses)}>"
            )
        return f"<{type(self).__name__}{type_str} stmt: {self.stmt.name}, uses: {len(self.uses)}>"

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_resultvalue(self)

    @classmethod
    def deserialize(
        cls: type[Self], serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> Self:
        return cast(Self, deserializer.deserialize_resultvalue(serUnit))


@dataclass
class BlockArgument(SSAValue):
    """SSAValue that is an argument to a [`Block`][kirin.ir.Block]."""

    block: Block = field(init=False)
    """The block that this argument belongs to."""
    index: int = field(init=False)
    """The index of this argument in the block's argument list."""

    def __init__(
        self, block: Block, index: int, type: TypeAttribute = AnyType()
    ) -> None:
        super().__init__()
        self.type = type
        self.block = block
        self.index = index

    @property
    def owner(self) -> Block:
        return self.block

    def delete(self, safe: bool = True) -> None:
        self.block.args.delete(self, safe=safe)

    def __hash__(self) -> int:
        return id(self)

    def __repr__(self) -> str:
        if self.name:
            return f"<{type(self).__name__}[{self.type}] {self.name}, uses: {len(self.uses)}>"
        return f"<{type(self).__name__}[{self.type}] index: {self.index}, uses: {len(self.uses)}>"

    def print_impl(self, printer: Printer) -> None:
        super().print_impl(printer)
        if not isinstance(self.type, AnyType):
            with printer.rich(style="comment"):
                printer.plain_print(" : ")
                printer.print(self.type)

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_blockargument(self)

    @classmethod
    def deserialize(
        cls: type[Self], serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> Self:
        return cast(Self, deserializer.deserialize_blockargument(serUnit))


@dataclass
class DeletedSSAValue(SSAValue):
    value: SSAValue = field(init=False)

    def __init__(self, value: SSAValue) -> None:
        super().__init__()
        self.value = value
        self.type = value.type

    def __hash__(self) -> int:
        return id(self)

    def __repr__(self) -> str:
        return f"<{type(self).__name__}[{self.type}] value: {self.value}, uses: {len(self.uses)}>"

    @property
    def owner(self) -> Statement | Block:
        return self.value.owner


@dataclass
class TestValue(SSAValue):
    """Test SSAValue for testing IR construction."""

    def __init__(self, type: TypeAttribute = AnyType()) -> None:
        super().__init__()
        self.type = type

    def __hash__(self) -> int:
        return id(self)

    @property
    def owner(self) -> Statement | Block:
        raise NotImplementedError



"""The indexing dialect for Python.

This module contains the dialect for the Python indexing syntax, including:

- The `GetItem` statement class.
- A base class `Subscript` for indexing statements.
- A trait `GetItemLike` for indexing statements.
- The lowering pass for the indexing statement.
- The concrete implementation of the indexing statement.
- The constant propagation implementation (special case) of the indexing statement.
- The type inference implementation of the indexing statement.
- A canonical rewrite rule for the rewriting of a given getitem-like
    statement to another getitem-like statement.
"""

import ast
from abc import abstractmethod
from typing import Generic, TypeVar
from dataclasses import dataclass

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.analysis import const
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.analysis.typeinfer import TypeInference

dialect = ir.Dialect("py.indexing")

GetItemLikeStmt = TypeVar("GetItemLikeStmt", bound=ir.Statement)


@dataclass(frozen=True, eq=False)
class GetItemLike(ir.Trait[ir.Statement], Generic[GetItemLikeStmt]):

    @abstractmethod
    def get_object(self, stmt: GetItemLikeStmt) -> ir.SSAValue: ...

    @abstractmethod
    def get_index(self, stmt: GetItemLikeStmt) -> ir.SSAValue: ...

    @abstractmethod
    def new(
        self, stmt_type: type[GetItemLikeStmt], obj: ir.SSAValue, index: ir.SSAValue
    ) -> GetItemLikeStmt: ...


PyGetItemLikeStmt = TypeVar("PyGetItemLikeStmt", bound="GetItem")


@dataclass(frozen=True, eq=False)
class PyGetItemLike(GetItemLike[PyGetItemLikeStmt]):

    def get_object(self, stmt: PyGetItemLikeStmt) -> ir.SSAValue:
        return stmt.obj

    def get_index(self, stmt: PyGetItemLikeStmt) -> ir.SSAValue:
        return stmt.index

    def new(
        self, stmt_type: type[PyGetItemLikeStmt], obj: ir.SSAValue, index: ir.SSAValue
    ) -> PyGetItemLikeStmt:
        return stmt_type(obj=obj, index=index)


# NOTE: in IR setindex is very different from getindex
# taking Julia's semantics as reference here
@statement
class Subscript(ir.Statement):
    pass


@statement(dialect=dialect)
class GetItem(Subscript):
    name = "getitem"
    traits = frozenset({ir.Pure(), PyGetItemLike(), lowering.FromPythonCall()})
    obj: ir.SSAValue = info.argument(print=False)
    index: ir.SSAValue = info.argument(print=False)
    result: ir.ResultValue = info.result(types.Any)


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_Subscript(
        self, state: lowering.State, node: ast.Subscript
    ) -> lowering.Result:
        value = state.lower(node.value).expect_one()
        slice = state.lower(node.slice).expect_one()
        if isinstance(node.ctx, ast.Load):
            stmt = GetItem(obj=value, index=slice)
        else:
            raise lowering.BuildError(f"unsupported subscript context {node.ctx}")
        return state.current_frame.push(stmt)


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(GetItem)
    def getindex(self, interp, frame: interp.Frame, stmt: GetItem):
        return (frame.get(stmt.obj)[frame.get(stmt.index)],)


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(GetItem)
    def getitem(
        self,
        interp: TypeInference,
        frame: interp.Frame[types.TypeAttribute],
        stmt: GetItem,
    ):
        obj = frame.get(stmt.obj)
        index: types.TypeAttribute = frame.get(stmt.index)
        # TODO: replace this when we can multiple dispatch
        if obj.is_subseteq(types.Tuple):
            return self.getitem_tuple(interp, stmt, obj, index)
        elif obj.is_subseteq(types.String):
            return (types.String,)
        else:
            return (types.Any,)

    def getitem_tuple(
        self,
        interp,
        stmt: GetItem,
        obj: types.TypeAttribute,
        index: types.TypeAttribute,
    ):
        if isinstance(obj, types.Generic):
            if index.is_subseteq(types.Int):
                return self.getitem_tuple_index(interp, stmt, obj, index)
            elif index.is_subseteq(types.Slice):
                return self.getitem_tuple_slice(interp, stmt, obj, index)
            else:
                return (types.Bottom,)
        elif isinstance(obj, types.PyClass):
            return (types.Any,)
        else:
            return (types.Bottom,)

    def getitem_tuple_index(
        self,
        interp: TypeInference,
        stmt: GetItem,
        obj: types.Generic,
        index: types.TypeAttribute,
    ):
        if index_ := interp.maybe_const(stmt.index, int):
            if obj.vararg and (index_ >= len(obj.vars) or -len(obj.vars) <= index_ < 0):
                return (obj.vararg.typ,)
            elif obj.vars and (
                0 <= index_ < len(obj.vars) or -len(obj.vars) <= index_ < 0
            ):
                return (obj.vars[index_],)
            else:
                return (types.Bottom,)
        else:
            return (self.getitem_tuple_union(obj),)

    def getitem_tuple_slice(
        self,
        interp: TypeInference,
        stmt: GetItem,
        obj: types.Generic,
        index: types.TypeAttribute,
    ):
        if index_ := interp.maybe_const(stmt.index, slice):
            if obj.vararg and index_.stop >= len(obj.vars):
                return (
                    types.Union(
                        *obj.vars[slice(index_.start, len(obj.vars), index_.step)],
                        obj.vararg.typ,
                    ),
                )
            elif index_.stop is None or index_.stop < len(obj.vars):
                return (
                    types.Tuple.where(
                        obj.vars[slice(index_.start, index_.stop, index_.step)]
                    ),
                )
            else:  # out of bounds
                return (types.Bottom,)
        else:
            return (types.Tuple[types.Vararg(self.getitem_tuple_union(obj))],)

    def getitem_tuple_union(self, obj: types.Generic):
        if obj.vararg:
            return types.Union(*obj.vars, obj.vararg.typ)
        else:
            return types.Union(*obj.vars)


@dialect.register(key="constprop")
class ConstProp(interp.MethodTable):

    @interp.impl(GetItem)
    def getitem(
        self,
        _: const.Propagate,
        frame: const.Frame,
        stmt: GetItem,
    ) -> interp.StatementResult[const.Result]:
        obj = frame.get(stmt.obj)
        index = frame.get(stmt.index)
        if not isinstance(index, const.Value):
            return (const.Unknown(),)

        if isinstance(obj, const.Value):
            return (const.Value(obj.data[index.data]),)
        elif isinstance(obj, const.PartialTuple):
            obj = obj.data
            if isinstance(index.data, int) and 0 <= index.data < len(obj):
                return (obj[index.data],)
            elif isinstance(index.data, slice):
                return (const.PartialTuple(obj[index.data]),)
        return (const.Unknown(),)


GetItemLikeStmt = TypeVar("GetItemLikeStmt", bound=ir.Statement)


@dataclass(init=False)
class RewriteGetItem(RewriteRule, Generic[GetItemLikeStmt]):
    target_stmt_type: type[GetItemLikeStmt]
    obj_type: types.TypeAttribute
    getitem_like: GetItemLike[GetItemLikeStmt]

    def __init__(self, stmt_type: type[GetItemLikeStmt], obj_type: types.TypeAttribute):
        trait = stmt_type.get_trait(GetItemLike)
        if trait is None:
            raise ValueError(f"{stmt_type} does not have GetItemLike trait")

        self.obj_type = obj_type
        self.target_stmt_type = stmt_type
        self.getitem_like = trait

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, GetItem):
            return RewriteResult()

        if not node.obj.type.is_subseteq(self.obj_type):
            return RewriteResult()

        node.replace_by(
            self.getitem_like.new(self.target_stmt_type, node.obj, node.index)
        )
        return RewriteResult(has_done_something=True)



from kirin.prelude import basic
from kirin.dialects import ilist


def test_map_wrapper():

    @basic
    def add1(x: int):
        return x + 1

    @basic
    def map_wrap():
        return ilist.map(add1, range(5))

    out = map_wrap()
    assert isinstance(out, ilist.IList)
    assert out.data == [1, 2, 3, 4, 5]


def test_foldr_wrapper():

    @basic
    def add_fold(x: int, out: int):
        return out + x

    @basic
    def map_foldr():
        return ilist.foldr(add_fold, range(5), init=10)

    out = map_foldr()
    assert isinstance(out, int)
    assert out == 10 + 0 + 1 + 2 + 3 + 4


def test_foldl_wrapper():

    @basic
    def add_fold2(out: int, x: int):
        return out + x

    @basic
    def map_foldl():
        return ilist.foldr(add_fold2, range(5), init=10)

    out = map_foldl()
    assert isinstance(out, int)
    assert out == 10 + 0 + 1 + 2 + 3 + 4


def test_scan_wrapper():

    @basic
    def add_scan(out: int, x: int):
        return out + 1, out + x

    @basic
    def scan_wrap():
        return ilist.scan(add_scan, range(5), init=10)

    out = scan_wrap()
    assert isinstance(out, tuple)
    assert len(out) == 2

    res = out[0]
    out_list = out[1]

    assert isinstance(res, int)
    assert res == 10 + 1 * 5

    assert isinstance(out_list, ilist.IList)
    assert out_list.data == [
        10 + 0,
        10 + 1 + 1,
        10 + 1 + 1 + 2,
        10 + 1 + 1 + 1 + 3,
        10 + 1 + 1 + 1 + 1 + 4,
    ]


def test_any_all_wrapper():

    @basic
    def test_any_all():
        ls = [True, False, False]
        return ls, ilist.any(ls), ilist.all(ls)

    test_any_all.print()

    ls, any_val, all_val = test_any_all()

    assert isinstance(ls, ilist.IList)
    assert ls.data == [True, False, False]
    assert any_val
    assert not all_val

    @basic
    def test_any_all2():
        ls = [False, False]
        return ilist.any(ls), ilist.all(ls)

    any_val, all_val = test_any_all2()
    assert not any_val
    assert not all_val

    @basic
    def test_any_all3():
        ls = [True, True, True, True, True]
        return ilist.any(ls), ilist.all(ls)

    any_val, all_val = test_any_all3()
    assert any_val
    assert all_val


def test_sorted():
    def key_test(a: int) -> int:
        return a

    @basic
    def main():
        ls = [2, 3, 1, 5, 4]
        return (
            ilist.sorted(ls),
            ilist.sorted(ls, key=key_test),
            ilist.sorted(ls, reverse=True),
        )

    main.print()

    ls1, ls2, ls3 = main()
    assert ls1.data == [1, 2, 3, 4, 5]
    assert ls2.data == ls1.data
    assert ls3.data == [5, 4, 3, 2, 1]

    def first(x: tuple[str, int]) -> str:
        return x[0]

    def second(x: tuple[str, int]) -> int:
        return x[1]

    @basic
    def main2():
        ls = [("a", 4), ("b", 3), ("c", 1)]
        return (
            ilist.sorted(ls, key=first),
            ilist.sorted(ls, key=second),
            ilist.sorted(ls, key=second, reverse=True),
        )

    main2.print()

    ls1, ls2, ls3 = main2()
    assert ls1.data == [("a", 4), ("b", 3), ("c", 1)]
    assert ls3.data == ls1.data
    assert ls2.data == [("c", 1), ("b", 3), ("a", 4)]



from kirin import ir, types
from kirin.rewrite import Walk
from kirin.dialects.py import Add, Div, Sub, Mult, BinOp
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.ir.nodes.base import IRNode
from kirin.dialects.ilist import IListType

from ..stmts import add as vadd, div as vdiv, sub as vsub, mult as vmult
from .._dialect import dialect


class DesugarBinOp(RewriteRule):
    """
    Convert py.BinOp statements with one scalar arg and one IList arg
    to the corresponding vmath binop. Currently supported binops are
    add, mult, sub, and div. BinOps where both args are IList are not
    supported, since `+` between two IList objects is taken to mean
    concatenation.
    """

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        match node:
            case BinOp():
                if node.lhs.type.is_subseteq(types.Bottom) or node.rhs.type.is_subseteq(
                    types.Bottom
                ):
                    return RewriteResult()
                elif (
                    node.lhs.type.is_subseteq(types.Number)
                    and node.rhs.type.is_subseteq(IListType)
                ) or (
                    node.lhs.type.is_subseteq(IListType)
                    and node.rhs.type.is_subseteq(types.Number)
                ):
                    return self.replace_binop(node)

            case _:
                return RewriteResult()

        return RewriteResult()

    def replace_binop(self, node: ir.Statement) -> RewriteResult:
        match node:
            case Add():
                node.replace_by(vadd(lhs=node.lhs, rhs=node.rhs))
                return RewriteResult(has_done_something=True)
            case Sub():
                node.replace_by(vsub(lhs=node.lhs, rhs=node.rhs))
                return RewriteResult(has_done_something=True)
            case Mult():
                node.replace_by(vmult(lhs=node.lhs, rhs=node.rhs))
                return RewriteResult(has_done_something=True)
            case Div():
                node.replace_by(vdiv(lhs=node.lhs, rhs=node.rhs))
                return RewriteResult(has_done_something=True)
            case _:
                return RewriteResult()


@dialect.post_inference
class WalkDesugarBinop(RewriteRule):
    """
    Walks DesugarBinop. Needed for correct behavior when
    registering as a post-inference rewrite.
    """

    def rewrite(self, node: IRNode):
        return Walk(DesugarBinOp()).rewrite(node)



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar
from dataclasses import field, dataclass

from kirin import ir
from kirin.ir.exception import ValidationError, ValidationErrorGroup

T = TypeVar("T")


class ValidationPass(ABC, Generic[T]):
    """Base class for a validation pass.

    Each pass analyzes an IR method and collects validation errors.
    """

    @abstractmethod
    def name(self) -> str:
        """Return the name of this validation pass."""
        ...

    @abstractmethod
    def run(self, method: ir.Method) -> tuple[Any, list[ValidationError]]:
        """Run validation and return (analysis_frame, errors).

        Returns:
            - analysis_frame: The result frame from the analysis
            - errors: List of validation errors (empty if valid)
        """
        ...

    def get_required_analyses(self) -> list[type]:
        """Return list of analysis classes this pass depends on.

        Override to declare dependencies (e.g., [AddressAnalysis, AnotherAnalysis]).
        The suite will run these analyses once and cache results.
        """
        return []

    def set_analysis_cache(self, cache: dict[type, Any]) -> None:
        """Receive cached analysis results from the suite.

        Override to store cached analysis frames/results.
        Example:
            self._address_frame = cache.get(AddressAnalysis)
        """
        pass


@dataclass
class ValidationSuite:
    """Compose multiple validation passes and run them together.

    Caches analysis results to avoid redundant computation when multiple
    validation passes depend on the same underlying analysis.

    fail_fast: If True, stops at the first validation pass that fails.

    Example:
        suite = ValidationSuite([
            NoCloningValidation,
            AnotherValidation,
        ])
        result = suite.validate(my_kernel)
        result.raise_if_invalid()
    """

    passes: list[type[ValidationPass]] = field(default_factory=list)
    fail_fast: bool = False
    _analysis_cache: dict[type, Any] = field(default_factory=dict, init=False)

    def add_pass(self, pass_cls: type[ValidationPass]) -> "ValidationSuite":
        """Add a validation pass to the suite."""
        self.passes.append(pass_cls)
        return self

    def validate(self, method: ir.Method) -> "ValidationResult":
        """Run all validation passes and collect results."""
        all_errors: dict[str, list[ValidationError]] = {}
        all_frames: dict[str, Any] = {}
        self._analysis_cache.clear()

        for pass_cls in self.passes:
            validator = pass_cls()
            pass_name = validator.name()

            try:
                required = validator.get_required_analyses()
                for required_analysis in required:
                    if required_analysis not in self._analysis_cache:
                        analysis = required_analysis(method.dialects)
                        analysis.initialize()
                        frame, _ = analysis.run(method)
                        self._analysis_cache[required_analysis] = frame

                validator.set_analysis_cache(self._analysis_cache)

                frame, errors = validator.run(method)
                all_frames[pass_name] = frame

                for err in errors:
                    if isinstance(err, ValidationError):
                        try:
                            err.attach(method)
                        except Exception:
                            pass

                if errors:
                    all_errors[pass_name] = errors
                    if self.fail_fast:
                        break

            except Exception as e:
                import traceback

                tb = traceback.format_exc()
                all_errors[pass_name] = [
                    ValidationError(
                        method.code, f"Validation pass '{pass_name}' failed: {e}\n{tb}"
                    )
                ]
                if self.fail_fast:
                    break

        return ValidationResult(all_errors, all_frames)


@dataclass
class ValidationResult:
    """Result of running a validation suite."""

    errors: dict[str, list[ValidationError]]
    frames: dict[str, Any] = field(default_factory=dict)
    is_valid: bool = field(default=True, init=False)

    def __post_init__(self):
        for _, errors in self.errors.items():
            if errors:
                self.is_valid = False
                break

    def error_count(self) -> int:
        """Total number of violations across all passes.

        Counts violations directly from frames using the same logic as test helpers.
        """

        total = 0
        for pass_name, errors in self.errors.items():
            if errors is None:
                continue
            total += len(errors)
        return total

    def get_frame(self, pass_name: str) -> Any:
        """Get the analysis frame for a specific pass."""
        return self.frames.get(pass_name)

    def _format_errors(self) -> str:
        """Format all errors with their pass names."""
        if self.is_valid:
            return "\n\033[32mAll validation passes succeeded\033[0m"

        lines = [
            f"\n\033[31mValidation failed with {self.error_count()} violation(s):\033[0m"
        ]
        for pass_name, pass_errors in self.errors.items():
            lines.append(f"\n\033[31m{pass_name}:\033[0m")
            for err in pass_errors:
                err_msg = err.args[0] if err.args else str(err)
                lines.append(f"  - {err_msg}")
                if hasattr(err, "hint"):
                    hint = err.hint()
                    if hint:
                        lines.append(f"    {hint}")

        return "\n".join(lines)

    def raise_if_invalid(self):
        """Raise an exception if validation failed."""
        if not self.is_valid:
            exceptions = []
            for _, pass_errors in self.errors.items():
                exceptions.extend(pass_errors)

            message = self._format_errors()
            raise ValidationErrorGroup(message, errors=exceptions)



"""Kirin IR Traits.

This module defines the traits that can be used to define the behavior of
Kirin IR nodes. The base trait is `StmtTrait`, which is a `dataclass` that
implements the `__hash__` and `__eq__` methods.

There are also some basic traits that are provided for convenience, such as
`Pure`, `HasParent`, `ConstantLike`, `IsTerminator`, `NoTerminator`, and
`IsolatedFromAbove`.
"""

from .abc import (
    Trait as Trait,
    AttrTrait as AttrTrait,
    StmtTrait as StmtTrait,
    RegionGraph as RegionGraph,
    RegionInterpretationTrait as RegionInterpretationTrait,
)
from .basic import (
    Pure as Pure,
    HasParent as HasParent,
    MaybePure as MaybePure,
    ConstantLike as ConstantLike,
    IsTerminator as IsTerminator,
    NoTerminator as NoTerminator,
    IsolatedFromAbove as IsolatedFromAbove,
)
from .symbol import (
    SymbolTable as SymbolTable,
    SymbolOpInterface as SymbolOpInterface,
    EntryPointInterface as EntryPointInterface,
)
from .callable import (
    StaticCall as StaticCall,
    HasSignature as HasSignature,
    CallableStmtInterface as CallableStmtInterface,
)
from .region.ssacfg import SSACFG as SSACFG, HasCFG as HasCFG



from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar, ClassVar, Generator
from contextlib import contextmanager
from dataclasses import field, dataclass

from typing_extensions import Self

from kirin import ir
from kirin.exception import KIRIN_INTERP_STATE

from .frame import FrameABC
from .state import InterpreterState
from .table import Signature, BoundedDef
from .value import ReturnValue, RegionResult, SpecialValue, StatementResult
from .exceptions import InterpreterError, StackOverflowError

ValueType = TypeVar("ValueType")
FrameType = TypeVar("FrameType", bound=FrameABC)


@dataclass
class InterpreterABC(ABC, Generic[FrameType, ValueType]):
    keys: ClassVar[tuple[str, ...]]
    """The name of the interpreter to select from dialects by order.
    First matching key will be used.
    """

    void: ValueType = field(init=False)
    """What to return when the interpreter evaluates nothing.
    """

    dialects: ir.DialectGroup
    """The dialects this interpreter supports."""

    max_depth: int = field(default=800, kw_only=True)
    """The maximum depth of the interpreter stack."""
    max_python_recursion_depth: int = field(default=131072, kw_only=True)
    """The maximum recursion depth of the Python interpreter.
    """
    debug: bool = field(default=False, kw_only=True)
    """Enable debug mode."""

    registry: dict[Signature, BoundedDef] = field(init=False, compare=False)
    """The registry of implementations"""
    symbol_table: dict[str, ir.Statement] = field(init=False, compare=False)
    """The symbol table of the interpreter."""
    state: InterpreterState[FrameType] = field(init=False, compare=False)
    """The interpreter state."""
    __eval_lock: bool = field(default=False, init=False, repr=False)
    """Lock for the eval method."""
    _validation_errors: dict[ir.IRNode, set[ir.ValidationError]] = field(
        default_factory=dict, init=False
    )
    """The validation errors collected during interpretation."""

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if ABC in cls.__bases__:
            return

        if not hasattr(cls, "keys"):
            raise TypeError(f"keys is not defined for class {cls.__name__}")
        if not hasattr(cls, "void"):
            raise TypeError(f"void is not defined for class {cls.__name__}")

    def __post_init__(self) -> None:
        self.registry = self.dialects.registry.interpreter(keys=self.keys)
        self.symbol_table = self.dialects.symbol_table

    def initialize(self) -> Self:
        self.state = InterpreterState()
        return self

    @abstractmethod
    def initialize_frame(
        self, node: ir.Statement, *, has_parent_access: bool = False
    ) -> FrameType:
        """Initialize a new call frame for the given callable node."""
        ...

    def call(
        self, node: ir.Statement | ir.Method, *args: ValueType, **kwargs: ValueType
    ) -> tuple[FrameType, ValueType]:
        """Call a given callable node with the given arguments.

        This method is used to call a node that has a callable trait and a
        corresponding implementation of its callable region execution convention in
        the interpreter.

        Args:
            node: the callable node to call
            args: the arguments to pass to the callable node
            kwargs: the keyword arguments to pass to the callable node

        Returns:
            tuple[FrameType, ValueType]: the frame and the result of the call

        Raises:
            InterpreterError: if the interpreter is already evaluating
            StackOverflowError: if the maximum depth of the interpreter stack is reached
        """
        if isinstance(node, ir.Method):
            return self.__call_method(node, *args, **kwargs)

        with self.new_frame(node) as frame:
            return frame, self.frame_call(frame, node, *args, **kwargs)

    def eval(self, node: ir.Statement) -> tuple[FrameType, StatementResult[ValueType]]:
        with self.new_frame(node) as frame:
            return frame, self.frame_eval(frame, node)

    def __call_method(
        self, node: ir.Method, *args: ValueType, **kwargs: ValueType
    ) -> tuple[FrameType, ValueType]:
        if self.__eval_lock:
            raise InterpreterError(
                f"Interpreter {self.__class__.__name__} is already evaluating, "
                f"consider calling the bare `method.code` instead of the method"
            )

        if node.nargs != len(args) + len(kwargs):
            raise InterpreterError(
                f"Method {node} called with {len(args) + len(kwargs)} "
                f"arguments, expected {node.nargs}"
            )

        with self.eval_context():
            return self.call(node.code, *args, **kwargs)

    @contextmanager
    def eval_context(self):
        """Context manager to set the recursion limit and initialize the interpreter.

        This context manager sets the recursion limit to the maximum depth of
        the interpreter stack. It is used to prevent stack overflow when calling
        recursive functions.
        """
        if self.__eval_lock:
            raise InterpreterError(
                f"Interpreter {self.__class__.__name__} is already evaluating, "
                f"consider calling the bare `method.code` instead of the method"
                f" or use the bare `frame_call`/`frame_eval` methods"
            )

        self.__eval_lock = True
        self.initialize()
        current_recursion_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(self.max_python_recursion_depth)
        try:
            yield self.max_python_recursion_depth
        except Exception as e:
            # NOTE: insert the interpreter state into the exception
            # so we can print the stack trace
            setattr(e, KIRIN_INTERP_STATE, self.state)
            raise e
        finally:
            self.__eval_lock = False
            sys.setrecursionlimit(current_recursion_limit)

    def frame_call(
        self,
        frame: FrameType,
        node: ir.Statement,
        *args: ValueType,
        **kwargs: ValueType,
    ) -> ValueType:
        """Call a given callable node with the given arguments in a given frame.

        This method is used to call a node that has a callable trait and a
        corresponding implementation of its callable region execution convention in
        the interpreter.
        """
        if entry := node.get_trait(ir.EntryPointInterface):
            node = self.symbol_table[entry.get_entry_point_symbol(node)]
        trait = node.get_present_trait(ir.CallableStmtInterface)
        args = trait.align_input_args(node, *args, **kwargs)
        region = trait.get_callable_region(node)
        if self.state.depth >= self.max_depth:
            return self.recursion_limit_reached()

        ret = self.frame_call_region(frame, node, region, *args)
        if isinstance(ret, ReturnValue):
            return ret.value
        elif not ret:  # empty result or None
            return self.void
        raise InterpreterError(
            f"callable region {node.name} does not return `ReturnValue`, got {ret}"
        )

    def recursion_limit_reached(self) -> ValueType:
        """Handle the recursion limit reached.

        This method is called when the maximum depth of the interpreter stack
        when calling a callable node is reached. By default a `StackOverflowError`
        is raised. Overload this method to provide a custom behavior, e.g. in
        the case of abstract interpreter, the recursion limit returns a bottom
        value.
        """
        raise StackOverflowError(
            f"Interpreter {self.__class__.__name__} stack "
            f"overflow at {self.state.depth}"
        )

    def frame_call_region(
        self,
        frame: FrameType,
        node: ir.Statement,
        region: ir.Region,
        *args: ValueType,
    ) -> RegionResult:
        """Call a given callable region with the given arguments in a given frame.

        This method is used to call a region that has a callable trait and a
        corresponding implementation of its callable region execution convention in
        the interpreter.

        Args:
            frame: the frame to call the region in
            node: the node to call the region on
            region: the region to call
            args: the arguments to pass to the region

        Returns:
            RegionResult: the result of the call

        Raises:
            InterpreterError: if cannot find a matching implementation for the region.
        """
        region_trait = node.get_present_trait(ir.RegionInterpretationTrait)

        how = self.registry.get(Signature(region_trait))
        if how is None:

            raise InterpreterError(
                f"Interpreter {self.__class__.__name__} does not "
                f"support {node} using {region_trait} convention"
            )
        region_trait.set_region_input(frame, region, *args)
        return how(self, frame, region)

    @contextmanager
    def new_frame(
        self, node: ir.Statement, *, has_parent_access: bool = False
    ) -> Generator[FrameType, Any, None]:
        """Create a new frame for the given node.

        This method is used to create a new call frame for the given node. The
        frame is pushed on the stack and popped when the context manager
        is exited. The frame is initialized with the given node and the
        given arguments.

        Args:
            node: the node to create the frame for
            has_parent_access: if the frame has access to the parent frame entries
                (default: False)

        Returns:
            Generator[FrameType, Any, None]: the frame
        """
        frame = self.initialize_frame(node, has_parent_access=has_parent_access)
        self.state.push_frame(frame)
        try:
            yield frame
        finally:
            self.state.pop_frame()

    def frame_eval(
        self, frame: FrameType, node: ir.Statement
    ) -> StatementResult[ValueType]:
        """Run a statement within the current frame. This is the entry
        point of running a statement. It will look up the statement implementation
        in the dialect registry, or optionally call a fallback implementation.

        Args:
            frame: the current frame
            node: the statement to run

        Returns:
            StatementResult: the result of running the statement
        """
        method = self.lookup_registry(frame, node)
        if method is not None:
            results = method(self, frame, node)
            if self.debug and not isinstance(results, (tuple, SpecialValue)):
                raise InterpreterError(
                    f"method must return tuple or SpecialResult, got {results}"
                )
            return results
        elif node.dialect not in self.dialects:
            name = node.dialect.name if node.dialect else "None"
            dialects = ", ".join(d.name for d in self.dialects)
            raise InterpreterError(
                f"Interpreter {self.__class__.__name__} does not "
                f"support {node} using {name} dialect. "
                f"Expected {dialects}"
            )

        return self.eval_fallback(frame, node)

    def eval_fallback(
        self, frame: FrameType, node: ir.Statement
    ) -> StatementResult[ValueType]:
        """The fallback implementation of statements.

        This is called when no implementation is found for the statement.

        Args:
            frame: the current frame
            stmt: the statement to run

        Returns:
            StatementResult: the result of running the statement

        Note:
            Overload this method to provide a fallback implementation for statements.
        """
        raise NotImplementedError(
            f"Missing implementation for {type(node).__name__} at {node.source}"
        )

    def lookup_registry(
        self, frame: FrameType, node: ir.Statement
    ) -> BoundedDef | None:
        sig = self.build_signature(frame, node)
        if sig in self.registry:
            return self.registry[sig]
        elif (method := self.registry.get(Signature(type(node)))) is not None:
            return method
        else:
            return None

    def build_signature(self, frame: FrameType, node: ir.Statement) -> Signature:
        return Signature(node.__class__, tuple(arg.type for arg in node.args))

    def add_validation_error(self, node: ir.IRNode, error: ir.ValidationError) -> None:
        """Add a ValidationError for a given IR node.

        If the node is not present in the _validation_errors dict, create a new set.
        Otherwise append to the existing set of errors.
        """
        self._validation_errors.setdefault(node, set()).add(error)

    def get_validation_errors(
        self, keys: set[ir.IRNode] | None = None
    ) -> list[ir.ValidationError]:
        """Get the validation errors collected during interpretation.

        If keys is provided, only return errors for the given nodes.
        Otherwise return all errors.
        """
        if keys is None:
            return [err for s in self._validation_errors.values() for err in s]
        return [
            err for node in keys for err in self._validation_errors.get(node, set())
        ]



from kirin.ir import Method
from kirin.interp import Frame, MethodTable, ReturnValue, impl, concrete

from .stmts import Call, Invoke, Lambda, Return, GetField, ConstantNone
from ._dialect import dialect


@dialect.register
class Interpreter(MethodTable):

    @impl(Call)
    def call(self, interp: concrete.Interpreter, frame: Frame, stmt: Call):
        mt: Method = frame.get(stmt.callee)
        _, ret = interp.call(
            mt.code,
            mt,
            *frame.get_values(stmt.inputs),
            **{k: v for k, v in zip(stmt.keys, frame.get_values(stmt.kwargs))},
        )
        return (ret,)

    @impl(Invoke)
    def invoke(self, interp: concrete.Interpreter, frame: Frame, stmt: Invoke):
        _, ret = interp.call(
            stmt.callee.code,
            stmt.callee,
            *frame.get_values(stmt.inputs),
        )
        return (ret,)

    @impl(Return)
    def return_(self, interp: concrete.Interpreter, frame: Frame, stmt: Return):
        return ReturnValue(frame.get(stmt.value))

    @impl(ConstantNone)
    def const_none(
        self, interp: concrete.Interpreter, frame: Frame, stmt: ConstantNone
    ):
        return (None,)

    @impl(GetField)
    def getfield(self, interp: concrete.Interpreter, frame: Frame, stmt: GetField):
        mt: Method = frame.get(stmt.obj)
        return (mt.fields[stmt.field],)

    @impl(Lambda)
    def lambda_(self, interp: concrete.Interpreter, frame: Frame, stmt: Lambda):
        return (
            Method(
                dialects=interp.dialects,
                code=stmt,
                nargs=len(stmt.body.blocks[0].args),
                arg_names=[
                    arg.name or f"%{idx}"
                    for idx, arg in enumerate(stmt.body.blocks[0].args)
                ],
                fields=frame.get_values(stmt.captured),
            ),
        )



import sys
import inspect
from typing import Any, TypedDict

from typing_extensions import Unpack, Optional

from kirin.ir import Dialect
from kirin.decl.info import StatementFields


class StatementOptions(TypedDict, total=False):
    init: bool
    repr: bool
    kw_only: bool
    dialect: Optional[Dialect]
    property: bool


class BaseModifier:
    _PARAMS = "__kirin_stmt_params"

    def __init__(self, cls: type, **kwargs: Unpack[StatementOptions]) -> None:
        self.cls = cls
        self.cls_module = sys.modules.get(cls.__module__)

        if "dialect" in kwargs:
            self.dialect = kwargs["dialect"]
        else:
            self.dialect = None
        self.params = kwargs
        setattr(cls, self._PARAMS, self.params)

        if cls.__module__ in sys.modules:
            self.globals = sys.modules[cls.__module__].__dict__
        else:
            # Theoretically this can happen if someone writes
            # a custom string to cls.__module__.  In which case
            # such dataclass won't be fully introspectable
            # (w.r.t. typing.get_type_hints) but will still function
            # correctly.
            self.globals: dict[str, Any] = {}

        # analysis state, used by scan_field, etc.
        self.fields = StatementFields()
        self.has_statement_bases = False
        self.kw_only = self.params.get("kw_only", False)
        self.KW_ONLY_seen = False

    def register(self) -> None:
        if self.dialect is None:
            return
        self.dialect.register(self.cls)

    def emit(self):
        self._self_name = "__kirin_stmt_self" if "self" in self.fields else "self"
        self._class_name = "__kirin_stmt_cls" if "cls" in self.fields else "cls"
        self._run_passes("emit_")

    def verify(self):
        self._run_passes("verify_")

    def _run_passes(self, prefix: str):
        for name, method in inspect.getmembers(self, inspect.ismethod):
            if name.startswith(prefix):
                method()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from kirin.serialization.base.context import MethodSymbolMeta
    from kirin.serialization.core.serializationunit import SerializationUnit


class SerializationModule:
    symbol_table: dict[str, "MethodSymbolMeta"]
    body: "SerializationUnit"

    def __init__(
        self, symbol_table: dict[str, "MethodSymbolMeta"], body: "SerializationUnit"
    ):
        self.symbol_table = symbol_table
        self.body = body



from __future__ import annotations

import typing
from types import ModuleType

# from typing import TYPE_CHECKING, Generic, TypeVar, Callable, ParamSpec
from dataclasses import field, dataclass

from typing_extensions import Self

from kirin.print.printer import Printer
from kirin.print.printable import Printable

if typing.TYPE_CHECKING:
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.deserializer import Deserializer
    from kirin.serialization.core.serializationunit import SerializationUnit

from .traits import (
    StaticCall,
    HasSignature,
    SymbolOpInterface,
    EntryPointInterface,
    CallableStmtInterface,
)
from .exception import ValidationError
from .nodes.stmt import Statement
from .attrs.types import FunctionType

if typing.TYPE_CHECKING:
    from kirin.ir.group import DialectGroup

Param = typing.ParamSpec("Param")
RetType = typing.TypeVar("RetType")


@dataclass
class Method(Printable, typing.Generic[Param, RetType]):
    dialects: "DialectGroup"  # own
    """The dialects that creates the method. This should be a DialectGroup."""
    code: Statement  # own, the corresponding IR, a func.func usually
    """The code of the method. This should be a statement with CallableStmtInterface trait."""
    nargs: int
    """The number of arguments of the method. 0 if no arguments."""
    mod: ModuleType | None = None  # ref
    """The module where the method is defined. None if no module."""
    py_func: typing.Callable[Param, RetType] | None = None  # ref
    """The original Python function. None if no Python function."""
    sym_name: str | None = None
    """The name of the method. None if no name."""
    arg_names: list[str] | None = None
    """The argument names of the callable statement. None if no keyword arguments allowed."""
    # values contained if closure
    fields: tuple = field(default_factory=tuple)  # own
    """values captured in the method if it is a closure."""
    file: str = ""
    """The file where the method is defined. Empty string if no file."""
    lineno_begin: int = 0
    """The line number where the method is defined. 0 if no line number."""
    inferred: bool = False
    """if typeinfer has been run on this method
    """
    backedges: set[Method] = field(init=False, repr=False)
    """Cache for the backedges. (who calls this method)"""
    run_passes: typing.Callable[[Method], None] | None = field(init=False, repr=False)

    def __init__(
        self,
        dialects: DialectGroup,
        code: Statement,
        *,
        nargs: int | None = None,
        mod: ModuleType | None = None,
        py_func: typing.Callable[Param, RetType] | None = None,
        sym_name: str | None = None,
        arg_names: list[str] | None = None,
        fields: tuple = (),
        file: str = "",
        lineno_begin: int = 0,
        inferred: bool = False,
    ):
        callable_node = code
        if entry_point := code.get_trait(EntryPointInterface):
            callable_node = entry_point.get_entry_point(code)

        trait = callable_node.get_present_trait(CallableStmtInterface)
        region = trait.get_callable_region(callable_node)
        if sym_name is None and (
            symbol_trait := callable_node.get_trait(SymbolOpInterface)
        ):
            sym_name = symbol_trait.get_sym_name(callable_node).data

        assert (
            len(region.blocks[0].args) > 0
        ), "Method must have at least have a self argument"

        self.dialects = dialects
        self.code = code
        self.nargs = nargs if nargs is not None else len(region.blocks[0].args)
        self.mod = mod
        self.py_func = py_func
        self.sym_name = sym_name
        self.arg_names = arg_names or [
            arg.name or f"arg{i}" for i, arg in enumerate(region.blocks[0].args)
        ]
        self.fields = fields
        self.file = file
        self.lineno_begin = lineno_begin
        self.inferred = inferred
        self.backedges = set()
        self.update_backedges()
        self.run_passes = None

    def __hash__(self) -> int:
        return id(self)

    def __call__(self, *args: Param.args, **kwargs: Param.kwargs) -> RetType:
        from kirin.interp.concrete import Interpreter

        if len(args) + len(kwargs) != self.nargs - 1:
            raise ValueError(
                f"Incorrect number of arguments, expected {self.nargs - 1}, got {len(args) + len(kwargs)}"
            )
        # NOTE: multi-return values will be wrapped in a tuple for Python
        interp = Interpreter(self.dialects)
        _, ret = interp.run(self, *args, **kwargs)
        return ret

    @property
    def args(self):
        """Return the arguments of the method. (excluding self)"""
        return tuple(arg for arg in self.callable_region.blocks[0].args[1:])

    @property
    def arg_types(self):
        """Return the types of the arguments of the method. (excluding self)"""
        return tuple(arg.type for arg in self.args)

    @property
    def self_type(self):
        """Return the type of the self argument of the method."""
        trait = self.code.get_present_trait(HasSignature)
        signature = trait.get_signature(self.code)
        return FunctionType(params_type=signature.inputs, return_type=signature.output)

    @property
    def callable_region(self):
        trait = self.code.get_present_trait(CallableStmtInterface)
        return trait.get_callable_region(self.code)

    @property
    def return_type(self):
        trait = self.code.get_present_trait(HasSignature)
        return trait.get_signature(self.code).output

    def __repr__(self) -> str:
        name = self.sym_name or "<lambda>"
        return f'Method("{name}")'

    def print_impl(self, printer: Printer) -> None:
        return printer.print(self.code)

    def similar(self, dialects: typing.Optional["DialectGroup"] = None):
        return Method(
            dialects=dialects or self.dialects,
            code=self.code.from_stmt(self.code, regions=[self.callable_region.clone()]),
            nargs=self.nargs,
            mod=self.mod,
            py_func=self.py_func,
            sym_name=self.sym_name,
            arg_names=self.arg_names,
            fields=self.fields,
            file=self.file,
            lineno_begin=self.lineno_begin,
            inferred=self.inferred,
        )

    def verify(self) -> None:
        """verify the method body.

        This will raise a ValidationError if the method body is not valid.
        """
        try:
            self.code.verify()
        except ValidationError as e:
            e.attach(self)
            raise e

    def verify_type(self) -> None:
        """verify the method type.

        This will raise a ValidationError if the method type is not valid.
        """
        # NOTE: verify the method body
        self.verify()

        try:
            self.code.verify_type()
        except ValidationError as e:
            e.attach(self)
            raise e

    def update_backedges(self):
        """Update the backedges of callee methods. (if they are static calls)"""
        for stmt in self.code.walk():
            trait = stmt.get_trait(StaticCall)
            if not trait:
                continue

            callee = trait.get_callee(stmt)
            callee.backedges.add(self)

    def is_structurally_equal(self, other: Method, context: dict | None = None) -> bool:
        return (
            isinstance(other, Method)
            and self.dialects.is_structurally_equal(other.dialects)
            and self.sym_name == other.sym_name
            and self.arg_names == other.arg_names
            and self.arg_types == other.arg_types
            and self.return_type == other.return_type
            and self.code.is_structurally_equal(other.code, context)
        )

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_method(self)

    @classmethod
    def deserialize(
        cls: type[Self],
        serUnit: "SerializationUnit",
        deserializer: "Deserializer",
    ) -> Self:
        return typing.cast(Self, deserializer.deserialize_method(serUnit))



from kirin import ir, rewrite
from kirin.prelude import basic
from kirin.dialects import py, func
from kirin.dialects.func.rewrite import lambdalifting


def test_rewrite_inner_lambda():
    @basic
    def outer():
        def inner(x: int):
            return x + 1

        return inner

    pyconstant_stmt = outer.code.regions[0].blocks[0].stmts.at(0)
    assert isinstance(pyconstant_stmt, py.Constant), "expected a Constant in outer body"
    assert isinstance(
        pyconstant_stmt.value, ir.PyAttr
    ), "expected a PyAttr in outer body"
    assert isinstance(
        pyconstant_stmt.value.data.code, func.Lambda
    ), "expected a lambda Method in outer body"

    rewrite.Walk(lambdalifting.LambdaLifting()).rewrite(outer.code)
    assert isinstance(
        pyconstant_stmt.value.data.code, func.Function
    ), "expected a Function in outer body"


def test_rewrite_inner_lambda_with_captured_vars():
    @basic
    def outer2():
        z = 10
        y = 3 + z

        def inner2(x: int):
            return x + y + 5

        return inner2

    pyconstant_stmt = outer2.code.regions[0].blocks[0].stmts.at(0)
    assert isinstance(pyconstant_stmt, py.Constant), "expected a Constant in outer body"
    assert isinstance(
        pyconstant_stmt.value, ir.PyAttr
    ), "expected a PyAttr in outer body"
    assert isinstance(
        pyconstant_stmt.value.data.code, func.Lambda
    ), "expected a lambda Method in outer body"
    rewrite.Walk(lambdalifting.LambdaLifting()).rewrite(outer2.code)
    assert isinstance(
        pyconstant_stmt.value.data.code, func.Function
    ), "expected a Function in outer body"



from __future__ import annotations

from typing import IO, TypeVar

from kirin import emit, interp

from .stmts import Invoke, Return, Function
from ._dialect import dialect


@dialect.register(key="emit.julia")
class Julia(interp.MethodTable):

    IO_t = TypeVar("IO_t", bound=IO)

    @interp.impl(Return)
    def return_(
        self, emit: emit.Julia[IO_t], frame: emit.JuliaFrame[IO_t], node: Return
    ):
        value = frame.get(node.value)
        frame.write_line(f"return {value}")

    @interp.impl(Invoke)
    def invoke(
        self, emit: emit.Julia[IO_t], frame: emit.JuliaFrame[IO_t], node: Invoke
    ):
        func_name = emit.callables.get(node.callee.code)
        if func_name is None:
            emit.callable_to_emit.append(node.callee.code)
            func_name = emit.callables.add(node.callee.code)

        _, call_expr = emit.call(
            node.callee.code, func_name, *frame.get_values(node.args)
        )
        frame.write_line(f"{frame.ssa[node.result]} = {call_expr}")
        return (frame.ssa[node.result],)

    @interp.impl(Function)
    def function(
        self, emit_: emit.Julia[IO_t], frame: emit.JuliaFrame[IO_t], node: Function
    ):
        func_name = emit_.callables[node]
        frame.set(node.body.blocks[0].args[0], func_name)
        argnames_: list[str] = []
        for arg in node.body.blocks[0].args[1:]:
            frame.set(arg, name := frame.ssa[arg])
            argnames_.append(name)

        argnames = ", ".join(argnames_)
        frame.write_line(f"function {func_name}({argnames})")
        with frame.indent():
            for block in node.body.blocks:
                frame.current_block = block
                frame.write_line(f"@label {frame.block[block]}")
                for arg in block.args:
                    frame.set(arg, frame.ssa[arg])

                for stmt in block.stmts:
                    frame.current_stmt = stmt
                    stmt_results = emit_.frame_eval(frame, stmt)

                    match stmt_results:
                        case tuple():
                            frame.set_values(stmt._results, stmt_results)
                        case _:
                            continue
        frame.write_line("end\n")



import numpy as np
from scipy import special

from kirin import types
from kirin.prelude import basic
from kirin.dialects import ilist, vmath


@basic.union([vmath])
def add_kernel(x, y):
    return vmath.add(x, y)


def test_add_lists():
    a = ilist.IList([0.0, 1.0, 2.0], elem=types.Float)
    b = ilist.IList([3.0, 4.0, 5.0], elem=types.Float)
    truth = np.array([0.0, 1.0, 2.0]) + np.array([3.0, 4.0, 5.0])
    out = add_kernel(a, b)
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def sub_kernel(x, y):
    return vmath.sub(x, y)


def test_sub_lists():
    a = ilist.IList([5.0, 7.0, 9.0], elem=types.Float)
    b = ilist.IList([3.0, 4.0, 5.0], elem=types.Float)
    truth = np.array([5.0, 7.0, 9.0]) - np.array([3.0, 4.0, 5.0])
    out = sub_kernel(a, b)
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


def test_sub_scalar_list():
    a = 10.0
    b = ilist.IList([3.0, 4.0, 5.0], elem=types.Float)
    truth = 10.0 - np.array([3.0, 4.0, 5.0])
    out = sub_kernel(a, b)
    out2 = sub_kernel(b, a)

    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)

    truth2 = np.array([3.0, 4.0, 5.0]) - 10.0
    assert isinstance(out2, ilist.IList)
    assert out2.elem == types.Float
    assert np.allclose(out2, truth2)


def test_add_scalar_list():
    a = 2.0
    b = ilist.IList([3.0, 4.0, 5.0], elem=types.Float)
    truth = 2.0 + np.array([3.0, 4.0, 5.0])
    out = add_kernel(a, b)
    out2 = add_kernel(b, a)

    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)

    assert isinstance(out2, ilist.IList)
    assert out2.elem == types.Float
    assert np.allclose(out2, truth)


@basic.union([vmath])
def mult_kernel(x, y):
    return vmath.mult(x, y)


def test_mult_lists():
    a = ilist.IList([1.0, 2.0, 3.0], elem=types.Float)
    b = ilist.IList([4.0, 5.0, 6.0], elem=types.Float)
    truth = np.array([1.0, 2.0, 3.0]) * np.array([4.0, 5.0, 6.0])
    out = mult_kernel(a, b)
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


def test_mult_scalar_list():
    a = 3.0
    b = ilist.IList([4.0, 5.0, 6.0], elem=types.Float)
    truth = 3.0 * np.array([4.0, 5.0, 6.0])
    out = mult_kernel(a, b)
    out2 = mult_kernel(b, a)

    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)

    assert isinstance(out2, ilist.IList)
    assert out2.elem == types.Float
    assert np.allclose(out2, truth)


@basic.union([vmath])
def div_kernel(x, y):
    return vmath.div(x, y)


def test_div_lists():
    a = ilist.IList([8.0, 9.0, 10.0], elem=types.Float)
    b = ilist.IList([2.0, 3.0, 5.2], elem=types.Float)
    truth = np.array([8.0, 9.0, 10.0]) / np.array([2.0, 3.0, 5.2])
    out = div_kernel(a, b)
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


def test_div_scalar_list():
    a = 12.0
    b = ilist.IList([2.0, 3.0, 4.0], elem=types.Float)
    truth = 12.0 / np.array([2.0, 3.0, 4.0])
    out = div_kernel(a, b)
    out2 = div_kernel(b, a)

    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)

    truth2 = np.array([2.0, 3.0, 4.0]) / 12.0
    assert isinstance(out2, ilist.IList)
    assert out2.elem == types.Float
    assert np.allclose(out2, truth2)


@basic.union([vmath])
def acos_func(x):
    return vmath.acos(x)


def test_acos():
    truth = np.arccos(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = acos_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def asin_func(x):
    return vmath.asin(x)


def test_asin():
    truth = np.arcsin(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = asin_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def asinh_func(x):
    return vmath.asinh(x)


def test_asinh():
    truth = np.arcsinh(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = asinh_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def atan_func(x):
    return vmath.atan(x)


def test_atan():
    truth = np.arctan(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = atan_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def atan2_func(y, x):
    return vmath.atan2(y, x)


def test_atan2():
    truth = np.arctan2(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
    )
    out = atan2_func(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
    )
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def atanh_func(x):
    return vmath.atanh(x)


def test_atanh():
    truth = np.arctanh(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = atanh_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def ceil_func(x):
    return vmath.ceil(x)


def test_ceil():
    truth = np.ceil(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = ceil_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def copysign_func(x, y):
    return vmath.copysign(x, y)


def test_copysign():
    truth = np.copysign(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
    )
    out = copysign_func(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
    )
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def cos_func(x):
    return vmath.cos(x)


def test_cos():
    truth = np.cos(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = cos_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def cosh_func(x):
    return vmath.cosh(x)


def test_cosh():
    truth = np.cosh(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = cosh_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def degrees_func(x):
    return vmath.degrees(x)


def test_degrees():
    truth = np.degrees(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = degrees_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def erf_func(x):
    return vmath.erf(x)


def test_erf():
    truth = special.erf(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = erf_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def erfc_func(x):
    return vmath.erfc(x)


def test_erfc():
    truth = special.erfc(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = erfc_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def exp_func(x):
    return vmath.exp(x)


def test_exp():
    truth = np.exp(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = exp_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def expm1_func(x):
    return vmath.expm1(x)


def test_expm1():
    truth = np.expm1(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = expm1_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def fabs_func(x):
    return vmath.fabs(x)


def test_fabs():
    truth = np.fabs(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = fabs_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def floor_func(x):
    return vmath.floor(x)


def test_floor():
    truth = np.floor(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = floor_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def fmod_func(x, y):
    return vmath.fmod(x, y)


def test_fmod():
    truth = np.fmod(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
    )
    out = fmod_func(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
    )
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def gamma_func(x):
    return vmath.gamma(x)


def test_gamma():
    truth = special.gamma(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = gamma_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def isfinite_func(x):
    return vmath.isfinite(x)


def test_isfinite():
    truth = np.isfinite(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = isfinite_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Bool
    assert np.allclose(out, truth)


@basic.union([vmath])
def isinf_func(x):
    return vmath.isinf(x)


def test_isinf():
    truth = np.isinf(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = isinf_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Bool
    assert np.allclose(out, truth)


@basic.union([vmath])
def isnan_func(x):
    return vmath.isnan(x)


def test_isnan():
    truth = np.isnan(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = isnan_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Bool
    assert np.allclose(out, truth)


@basic.union([vmath])
def lgamma_func(x):
    return vmath.lgamma(x)


def test_lgamma():
    truth = special.loggamma(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = lgamma_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def log10_func(x):
    return vmath.log10(x)


def test_log10():
    truth = np.log10(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = log10_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def log1p_func(x):
    return vmath.log1p(x)


def test_log1p():
    truth = np.log1p(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = log1p_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def log2_func(x):
    return vmath.log2(x)


def test_log2():
    truth = np.log2(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = log2_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def pow_func(x, y):
    return vmath.pow(x, y)


def test_pow():
    truth = np.power(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        3.33,
    )
    out = pow_func(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        3.33,
    )
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def radians_func(x):
    return vmath.radians(x)


def test_radians():
    truth = np.radians(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = radians_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def remainder_func(x, y):
    return vmath.remainder(x, y)


def test_remainder():
    truth = np.remainder(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
    )
    out = remainder_func(
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
        ilist.IList([0.42, 0.87, 0.32], elem=types.Float),
    )
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def sin_func(x):
    return vmath.sin(x)


def test_sin():
    truth = np.sin(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = sin_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def sinh_func(x):
    return vmath.sinh(x)


def test_sinh():
    truth = np.sinh(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = sinh_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def sqrt_func(x):
    return vmath.sqrt(x)


def test_sqrt():
    truth = np.sqrt(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = sqrt_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def tan_func(x):
    return vmath.tan(x)


def test_tan():
    truth = np.tan(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = tan_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def tanh_func(x):
    return vmath.tanh(x)


def test_tanh():
    truth = np.tanh(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = tanh_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def trunc_func(x):
    return vmath.trunc(x)


def test_trunc():
    truth = np.trunc(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    out = trunc_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float))
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def scale_func(x, y):
    return vmath.scale(value=y, x=x)


def test_scale():
    a = 3.3
    truth = np.array(ilist.IList([0.42, 0.87, 0.32], elem=types.Float)) * a
    out = scale_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float), a)
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)


@basic.union([vmath])
def offset_func(x, y):
    return vmath.offset(value=y, x=x)


def test_offset():
    a = 3.3
    truth = np.array(ilist.IList([0.42, 0.87, 0.32], elem=types.Float)) + a
    out = offset_func(ilist.IList([0.42, 0.87, 0.32], elem=types.Float), a)
    assert isinstance(out, ilist.IList)
    assert out.elem == types.Float
    assert np.allclose(out, truth)



from kirin.prelude import structural, structural_no_opt
from kirin.dialects import py, func, ilist
from kirin.passes.aggressive import UnrollScf


def test_unroll_scf():
    @structural
    def main(r: list[int], cond: bool):
        if cond:
            for i in range(4):
                tmp = r[-1]
                if i < 2:
                    tmp += i * 2
                else:
                    for j in range(4):
                        if i > j:
                            tmp += i + j
                        else:
                            tmp += i - j

                r.append(tmp)
        else:
            for i in range(4):
                r.append(i)
        return r

    UnrollScf(structural).fixpoint(main)

    num_adds = 0
    num_calls = 0

    for op in main.callable_region.walk():
        if isinstance(op, py.Add):
            num_adds += 1
        elif isinstance(op, func.Call):
            num_calls += 1

    assert num_adds == 10
    assert num_calls == 8


def test_dce_unroll_typeinfer():
    # NOTE: tests bug in typeinfer preventing DCE from issue#564

    @structural_no_opt
    def main():
        ls = [1, 2, 3]
        for i in range(1):
            ls[i] = 10
        return ls

    UnrollScf(main.dialects).fixpoint(main)

    for stmt in main.callable_region.stmts():
        if isinstance(stmt, py.Constant) and isinstance(
            value := stmt.value, ilist.IList
        ):
            assert not isinstance(value.data, range), "Unused range not eliminated!"



"""The list dialect for Python.

This module contains the dialect for list semantics in Python, including:

- The `New` and `Append` statement classes.
- The lowering pass for list operations.
- The concrete implementation of list operations.
- The type inference implementation of list operations.

This dialect maps `list()`, `ast.List` and `append()` calls to the `New` and `Append` statements.
"""

from . import interp as interp, lowering as lowering, typeinfer as typeinfer
from .stmts import New as New, Append as Append
from ._dialect import dialect as dialect



import ast
from dataclasses import dataclass

from kirin import types, lowering
from kirin.dialects import py

from . import stmts as ilist
from ._dialect import dialect


@dialect.register
class PythonLowering(lowering.FromPythonAST):

    def lower_List(self, state: lowering.State, node: ast.List) -> lowering.Result:
        elts = tuple(state.lower(each).expect_one() for each in node.elts)

        if len(elts):
            typ = elts[0].type
            for each in elts:
                typ = typ.join(each.type)
        else:
            typ = types.Any

        return state.current_frame.push(ilist.New(values=tuple(elts), elem_type=typ))

    @lowering.akin(ilist.IList)
    def lower_Call_IList(
        self, state: lowering.State, node: ast.Call
    ) -> lowering.Result:
        if len(node.args) != 1:
            raise lowering.BuildError("IList constructor only takes one argument")
        value = node.args[0]
        if not isinstance(value, ast.List):
            raise lowering.BuildError("IList constructor only takes a list")

        if len(node.keywords) > 1:
            raise lowering.BuildError(
                "IList constructor only takes one keyword argument"
            )

        if len(node.keywords) == 1:
            kw = node.keywords[0]
            if kw.arg != "elem":
                raise lowering.BuildError(
                    "IList constructor only takes elem keyword argument"
                )
            elem = self.get_hint(state, kw.value)
            elts = tuple(state.lower(each).expect_one() for each in value.elts)
            stmt = ilist.New(values=tuple(elts), elem_type=elem)
            return state.current_frame.push(stmt)
        else:
            return self.lower_List(state, value)


@dataclass(frozen=True)
class SortedLowering(lowering.FromPythonCall["ilist.Sorted"]):
    """
    Custom lowering for Sorted to deal with optional arguments `key=None` and `reverse=False`
    """

    def lower(
        self, stmt: type["ilist.Sorted"], state: lowering.State[ast.AST], node: ast.Call
    ) -> lowering.Result:
        args = node.args

        if len(args) != 1:
            raise lowering.BuildError(
                f"Expected single argument in sorted, got {len(args)}"
            )
        collection = state.lower(args[0]).expect_one()

        key = None
        reverse = None
        for kwarg in node.keywords:
            if kwarg.arg == "key":
                key = state.lower(kwarg.value).expect_one()
            elif kwarg.arg == "reverse":
                reverse = state.lower(kwarg.value).expect_one()
            else:
                raise lowering.BuildError(
                    f"Got unexpected keyword argument in sorted {kwarg.arg}"
                )

        if key is None:
            key = state.current_frame.push(py.Constant(None)).result

        if reverse is None:
            reverse = state.current_frame.push(py.Constant(False)).result

        return state.current_frame.push(
            stmt(collection=collection, key=key, reverse=reverse)
        )



from __future__ import annotations

from kirin import ir, types
from kirin.interp import Frame, MethodTable, ReturnValue, impl
from kirin.analysis import const
from kirin.analysis.typeinfer import TypeInference
from kirin.dialects.func.stmts import (
    Call,
    Invoke,
    Lambda,
    Return,
    GetField,
    ConstantNone,
)
from kirin.dialects.func._dialect import dialect


# NOTE: a lot of the type infer rules are same as the builtin dialect
@dialect.register(key="typeinfer")
class TypeInfer(MethodTable):

    @impl(ConstantNone)
    def const_none(self, interp: TypeInference, frame: Frame, stmt: ConstantNone):
        return (types.NoneType,)

    @impl(Return)
    def return_(
        self, interp: TypeInference, frame: Frame[types.TypeAttribute], stmt: Return
    ) -> ReturnValue:
        if (
            isinstance(hint := stmt.value.hints.get("const"), const.Value)
            and hint.data is not None
        ):
            return ReturnValue(types.Literal(hint.data, frame.get(stmt.value)))
        return ReturnValue(frame.get(stmt.value))

    @impl(Call)
    def call(self, interp_: TypeInference, frame: Frame, stmt: Call):
        # give up on dynamic method calls
        mt = interp_.maybe_const(stmt.callee, ir.Method)
        if mt is None:  # not a constant method
            return self._solve_method_type(interp_, frame, stmt)

        if mt.inferred:  # so we don't end up in infinite loop
            return (mt.return_type,)

        _, ret = interp_.call(
            mt.code,
            interp_.method_self(mt),
            *frame.get_values(stmt.inputs),
            **{k: v for k, v in zip(stmt.keys, frame.get_values(stmt.kwargs))},
        )
        return (ret,)

    def _solve_method_type(self, interp: TypeInference, frame: Frame, stmt: Call):
        mt_inferred = frame.get(stmt.callee)

        if not isinstance(mt_inferred, types.FunctionType):
            return (types.Bottom,)

        return (mt_inferred.return_type,)

    @impl(Invoke)
    def invoke(self, interp_: TypeInference, frame: Frame, stmt: Invoke):
        if stmt.callee.inferred:  # so we don't end up in infinite loop
            return (stmt.callee.return_type,)

        _, ret = interp_.call(
            stmt.callee.code,
            interp_.method_self(stmt.callee),
            *frame.get_values(stmt.inputs),
        )
        return (ret,)

    @impl(Lambda)
    def lambda_(
        self, interp_: TypeInference, frame: Frame[types.TypeAttribute], stmt: Lambda
    ):
        body_frame, ret = interp_.call(
            stmt,
            types.MethodType,
            *tuple(arg.type for arg in stmt.body.blocks[0].args[1:]),
        )
        argtypes = tuple(arg.type for arg in stmt.body.blocks[0].args[1:])
        ret = types.MethodType[[*argtypes], ret]
        frame.entries.update(body_frame.entries)  # pass results back to upper frame
        self_ = stmt.body.blocks[0].args[0]
        frame.set(self_, ret)
        return (ret,)

    @impl(GetField)
    def getfield(self, interp: TypeInference, frame, stmt: GetField):
        return (stmt.result.type,)



from dataclasses import dataclass

from kirin import ir
from kirin.analysis import const
from kirin.dialects import cf
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.py.constant import Constant


@dataclass
class ConstantFold(RewriteRule):

    def get_const(self, value: ir.SSAValue):
        ret = value.hints.get("const")

        if ret is not None and isinstance(ret, const.Value):
            return ret
        return None

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if node.has_trait(ir.ConstantLike):
            return RewriteResult()
        elif isinstance(node, cf.ConditionalBranch):
            return self.rewrite_cf_ConditionalBranch(node)

        if not self.is_pure(node):
            return RewriteResult()

        has_done_something = False
        for old_result in node.results:
            if (value := self.get_const(old_result)) is not None:
                if not old_result.uses:
                    continue
                stmt = Constant(value.data)
                stmt.insert_before(node)
                old_result.replace_by(stmt.result)
                stmt.result.hints["const"] = value
                if old_result.name:
                    stmt.result.name = old_result.name
                has_done_something = True
        return RewriteResult(has_done_something=has_done_something)

    def rewrite_cf_ConditionalBranch(self, node: cf.ConditionalBranch):
        if (value := self.get_const(node.cond)) is not None:
            if value.data is True:
                cf.Branch(
                    arguments=node.then_arguments,
                    successor=node.then_successor,
                ).insert_before(node)
            elif value.data is False:
                cf.Branch(
                    arguments=node.else_arguments,
                    successor=node.else_successor,
                ).insert_before(node)
            else:
                raise ValueError(f"Invalid constant value for branch: {value.data}")
            node.delete()
            return RewriteResult(has_done_something=True)
        return RewriteResult()



from typing import Iterable
from dataclasses import field, dataclass

from kirin import ir
from kirin.print import Printable
from kirin.dialects import func
from kirin.print.printer import Printer


@dataclass
class CallGraph(Printable):
    """Call graph for a given [`ir.Method`][kirin.ir.Method].

    This class implements the [`kirin.graph.Graph`][kirin.graph.Graph] protocol.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    edges: dict[ir.Method, set[ir.Method]] = field(default_factory=dict)
    """Mapping from symbol names to edges (caller -> callee)."""
    backedges: dict[ir.Method, set[ir.Method]] = field(default_factory=dict)
    """Mapping from symbol names to backedges (callee -> caller)."""

    def __init__(self, mt: ir.Method):
        self.defs = {}
        self.edges = {}
        self.backedges = {}
        self.__build(mt)

    def __build(self, mt: ir.Method):
        for stmt in mt.callable_region.walk():
            if isinstance(stmt, func.Invoke):
                edges = self.edges.setdefault(stmt.callee, set())
                edges.add(mt)
                self.__build(stmt.callee)

        for caller in self.edges:
            for callee in self.edges[caller]:
                backedges = self.backedges.setdefault(callee, set())
                backedges.add(caller)

    def get_neighbors(self, node: ir.Method) -> Iterable[ir.Method]:
        """Get the neighbors of a node in the call graph."""
        return self.edges.get(node, ())

    def get_edges(self) -> Iterable[tuple[ir.Method, ir.Method]]:
        """Get the edges of the call graph."""
        for node, neighbors in self.edges.items():
            for neighbor in neighbors:
                yield node, neighbor

    def get_nodes(self) -> Iterable[ir.Method]:
        """Get the nodes of the call graph."""
        return self.edges.keys()

    def print_impl(self, printer: Printer) -> None:
        for idx, (caller, callee) in enumerate(self.edges.items()):
            printer.plain_print(caller)
            printer.plain_print(" -> ")
            printer.print_seq(
                callee, delim=", ", prefix="[", suffix="]", emit=printer.plain_print
            )
            if idx < len(self.edges) - 1:
                printer.print_newline()



from kirin.worklist import WorkList


def test_worklist():
    wl = WorkList()

    assert wl.is_empty()
    assert not wl
    assert len(wl) == 0

    assert wl.pop() is None

    assert wl.is_empty()
    assert not wl
    assert len(wl) == 0

    wl.append("A")

    assert not wl.is_empty()
    assert wl
    assert len(wl) == 1

    assert wl.pop() == "A"

    assert wl.is_empty()
    assert not wl
    assert len(wl) == 0

    wl.append("Z")
    wl.extend("BCGFEDCB")

    assert not wl.is_empty()
    assert wl
    assert len(wl) == 9

    assert wl.pop() == "Z"
    assert wl.pop() == "B"

    assert not wl.is_empty()
    assert wl
    assert len(wl) == 7

    rest = []
    while wl:
        rest.append(wl.pop())
    assert rest == list("CGFEDCB")

    assert wl.is_empty()
    assert not wl
    assert len(wl) == 0



import math as pymath
from typing import TypeVar

from kirin import lowering
from kirin.dialects import ilist

from . import stmts as stmts, interp as interp
from ._dialect import dialect as dialect
from .rewrites import desugar as desugar

pi = pymath.pi
e = pymath.e
tau = pymath.tau

ListLen = TypeVar("ListLen")


@lowering.wraps(stmts.add)
def add(
    lhs: ilist.IList[float, ListLen] | float,
    rhs: ilist.IList[float, ListLen] | float,
) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.acos)
def acos(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.asin)
def asin(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.asinh)
def asinh(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.atan)
def atan(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.atan2)
def atan2(
    y: ilist.IList[float, ListLen], x: ilist.IList[float, ListLen]
) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.atanh)
def atanh(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.ceil)
def ceil(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.copysign)
def copysign(
    x: ilist.IList[float, ListLen], y: ilist.IList[float, ListLen]
) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.cos)
def cos(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.cosh)
def cosh(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.degrees)
def degrees(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.div)
def div(
    lhs: ilist.IList[float, ListLen] | float,
    rhs: ilist.IList[float, ListLen] | float,
) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.erf)
def erf(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.erfc)
def erfc(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.exp)
def exp(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.expm1)
def expm1(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.fabs)
def fabs(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.floor)
def floor(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.fmod)
def fmod(
    x: ilist.IList[float, ListLen], y: ilist.IList[float, ListLen]
) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.gamma)
def gamma(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.isfinite)
def isfinite(x: ilist.IList[float, ListLen]) -> ilist.IList[bool, ListLen]: ...


@lowering.wraps(stmts.isinf)
def isinf(x: ilist.IList[float, ListLen]) -> ilist.IList[bool, ListLen]: ...


@lowering.wraps(stmts.isnan)
def isnan(x: ilist.IList[float, ListLen]) -> ilist.IList[bool, ListLen]: ...


@lowering.wraps(stmts.lgamma)
def lgamma(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.log10)
def log10(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.log1p)
def log1p(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.log2)
def log2(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.mult)
def mult(
    lhs: ilist.IList[float, ListLen] | float,
    rhs: ilist.IList[float, ListLen] | float,
) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.pow)
def pow(x: ilist.IList[float, ListLen], y: float) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.radians)
def radians(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.remainder)
def remainder(
    x: ilist.IList[float, ListLen], y: ilist.IList[float, ListLen]
) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.sin)
def sin(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.sinh)
def sinh(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.sqrt)
def sqrt(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.sub)
def sub(
    lhs: ilist.IList[float, ListLen] | float,
    rhs: ilist.IList[float, ListLen] | float,
) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.tan)
def tan(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.tanh)
def tanh(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.trunc)
def trunc(x: ilist.IList[float, ListLen]) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.scale)
def scale(
    value: float, x: ilist.IList[float, ListLen]
) -> ilist.IList[float, ListLen]: ...


@lowering.wraps(stmts.offset)
def offset(
    value: float, x: ilist.IList[float, ListLen]
) -> ilist.IList[float, ListLen]: ...



from kirin import ir, types
from kirin.prelude import basic, basic_no_opt
from kirin.rewrite import Walk, Fixpoint
from kirin.dialects import cf, func
from kirin.dialects.py import cmp, binop
from kirin.analysis.cfg import CFG
from kirin.dialects.func import Lambda
from kirin.rewrite.inline import Inline
from kirin.rewrite.compactify import CFGCompactify, CompactifyRegion
from kirin.dialects.py.constant import Constant


@basic_no_opt
def foo(x: int):  # type: ignore
    def goo(y: int):
        return x + y

    return goo


def test_cfg_compactify():
    compactify = Walk(CFGCompactify())
    Fixpoint(compactify).rewrite(foo.code)
    assert len(foo.callable_region.blocks[0].stmts) == 2
    stmt = foo.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, Lambda)
    assert len(stmt.body.blocks[0].stmts) == 3
    assert len(stmt.body.blocks) == 1


@basic_no_opt
def my_func(x: int, y: int):
    def foo(a: int, b: int):
        return a + b + x + y

    return foo


@basic_no_opt
def my_main_test_cfg():
    a = 3
    b = 4
    c = my_func(1, 2)
    return c(a, b) * 4


def test_compactify_replace_block_arguments():
    Walk(Inline(heuristic=lambda x: True)).rewrite(my_main_test_cfg.code)
    compactify = Walk(CFGCompactify())
    Fixpoint(compactify).rewrite(my_main_test_cfg.code)
    stmt = my_main_test_cfg.callable_region.blocks[0].stmts.at(5)
    assert isinstance(stmt, func.Lambda)
    assert isinstance(stmt.captured[0].owner, Constant)
    assert stmt.captured[0].name == "x"
    assert isinstance(stmt.captured[1].owner, Constant)
    assert stmt.captured[1].name == "y"


def test_compactify_single_branch_block():
    region = ir.Region()
    region.blocks.append(ir.Block())
    region.blocks.append(ir.Block())
    region.blocks.append(ir.Block())
    region.blocks.append(ir.Block())
    region.blocks[0].args.append_from(types.Any, "self")
    x = region.blocks[0].args.append_from(types.Any, "x")
    const_0 = Constant(0)
    const_n = Constant(3)
    const_n.result.name = "n"
    cond = cmp.Eq(x, const_0.result)
    cond.result.name = "cond"
    region.blocks[0].stmts.append(const_0)
    region.blocks[0].stmts.append(const_n)
    region.blocks[0].stmts.append(cond)
    region.blocks[0].stmts.append(
        cf.ConditionalBranch(
            cond.result,
            then_arguments=(),
            then_successor=region.blocks[1],
            else_arguments=(),
            else_successor=region.blocks[2],
        )
    )
    region.blocks[1].stmts.append(
        cf.Branch(arguments=(const_n.result,), successor=region.blocks[3])
    )
    const_1 = Constant(1)
    sub = binop.Sub(x, const_1.result)
    region.blocks[2].stmts.append(const_1)
    region.blocks[2].stmts.append(sub)
    region.blocks[2].stmts.append(
        cf.Branch(arguments=(sub.result,), successor=region.blocks[3])
    )
    z = region.blocks[3].args.append_from(types.Any, "z")
    mul = binop.Mult(x, z)
    region.blocks[3].stmts.append(mul)
    region.blocks[3].stmts.append(func.Return(mul.result))
    cfg = CFG(region)
    compactify = Walk(CompactifyRegion(cfg))
    compactify.rewrite(region)
    region.print()
    assert len(region.blocks) == 3
    stmt = region.blocks[0].last_stmt
    assert isinstance(stmt, cf.ConditionalBranch)
    assert stmt.then_successor is region.blocks[2]


def test_compactify_entry_block_single_branch():
    region = ir.Region()
    region.blocks.append(ir.Block())
    region.blocks.append(ir.Block())
    region.blocks[0].args.append_from(types.Any, "self")
    x = region.blocks[0].args.append_from(types.Any, "x")
    region.blocks[0].stmts.append(cf.Branch(arguments=(), successor=region.blocks[1]))
    x_1 = Constant(0)
    x_2 = binop.Add(x, x_1.result)
    region.blocks[1].stmts.append(x_1)
    region.blocks[1].stmts.append(x_2)
    region.blocks[1].stmts.append(func.Return(x_2.result))
    cfg = CFG(region)
    compactify = Walk(CompactifyRegion(cfg))
    compactify.rewrite(region)

    target = ir.Region(ir.Block())
    target.blocks[0].args.append_from(types.Any, "self")
    x = target.blocks[0].args.append_from(types.Any, "x")
    x0 = Constant(0)
    target.blocks[0].stmts.append(x0)
    x1 = binop.Add(x, x0.result)
    target.blocks[0].stmts.append(x1)
    target.blocks[0].stmts.append(func.Return(x1.result))
    assert region.is_structurally_equal(target)


def test_compactify_dead_subgraph():
    @basic
    def deadblock_mwe():
        j = 0
        if False:
            j = 1

            if True:
                j = j + 1

            else:
                j = j - 1

        return j

    deadblock_mwe()



"""Tools for testing methods."""

from typing import List

from kirin.ir import Method, Statement
from kirin.dialects import func


def assert_statements_same(
    statement_1: Statement, statement_2: Statement, check_args: bool = True
):
    """Assert statements are the same and recursively check arguments.

    Args:
        statement_1 (Statement): First statement to compare.
        statement_2 (Statement): Second statement to compare.
        check_args (bool): Recursively check arguments if True.
    """
    assert type(statement_1) is type(
        statement_2
    ), "Statements have different type: {} vs {}".format(
        type(statement_1).__name__, type(statement_2).__name__
    )
    assert statement_1.attributes == statement_2.attributes, (
        "Statements have different attributes" ""
    )
    assert len(statement_1.args.field) == len(
        statement_2.args.field
    ), "Arguments have different lengths"
    if check_args:
        for arg_1, arg_2 in zip(statement_1.args.field, statement_2.args.field):
            assert isinstance(arg_1.owner, Statement)
            assert isinstance(arg_2.owner, Statement)
            assert_statements_same(arg_1.owner, arg_2.owner, check_args=True)


def assert_structurally_same(
    method: Method,
    expected_stmts: List[Statement],
    check_args: bool = True,
    debug: bool = False,
):
    """Assert Method is structurally the same as list of Statements.

    Args:
        method (Method): A kirin Method.
        expected_stmts: A list of statements to compare with.
        check_args (bool): Recursively check arguments of statements if True.
        debug (bool): Pretty-print statements in method until fail.
    """

    assert isinstance(method.code, func.Function)
    new_stmts = list(method.code.body.blocks[0].stmts)
    assert len(new_stmts) == len(expected_stmts), "Methods different lengths"
    for new_stmt, expected_stmt in zip(new_stmts, expected_stmts):
        if debug:
            new_stmt.print()
        assert_statements_same(new_stmt, expected_stmt, check_args=check_args)



from __future__ import annotations

from typing import Any
from dataclasses import dataclass

from kirin import ir

from .abc import InterpreterABC
from .frame import Frame


@dataclass
class Interpreter(InterpreterABC[Frame[Any], Any]):
    keys = ("main",)
    void = None

    def initialize_frame(
        self, node: ir.Statement, *, has_parent_access: bool = False
    ) -> Frame[Any]:
        """Initialize the frame for the given node."""
        return Frame(node, has_parent_access=has_parent_access)

    def run(self, method: ir.Method, *args, **kwargs):
        return self.call(method, method, *args, **kwargs)



"""
Immutable list dialect for Python.

This dialect provides a simple, immutable list dialect similar
to Python's built-in list type.
"""

from . import (
    _julia as _julia,
    interp as interp,
    rewrite as rewrite,
    lowering as lowering,
    constprop as constprop,
    typeinfer as typeinfer,
)
from .stmts import (
    Map as Map,
    New as New,
    Push as Push,
    Scan as Scan,
    Foldl as Foldl,
    Foldr as Foldr,
    ForEach as ForEach,
    IListType as IListType,
)
from .passes import IListDesugar as IListDesugar
from .runtime import IList as IList
from ._dialect import dialect as dialect
from ._wrapper import (  # careful this is not the builtin range
    all as all,
    any as any,
    map as map,
    scan as scan,
    foldl as foldl,
    foldr as foldr,
    range as range,
    sorted as sorted,
    for_each as for_each,
)



from kirin import ir, types
from kirin.decl import info, statement
from kirin.print.printer import Printer
from kirin.dialects.cf.dialect import dialect


@statement(dialect=dialect)
class Branch(ir.Statement):
    name = "br"
    traits = frozenset({ir.IsTerminator()})

    arguments: tuple[ir.SSAValue, ...]
    successor: ir.Block = info.block()

    def verify(self) -> None:
        return

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="keyword"):
            printer.print_name(self)

        printer.plain_print(" ")
        printer.plain_print(printer.state.block_id[self.successor])
        printer.print_seq(
            self.arguments,
            delim=", ",
            prefix="(",
            suffix=")",
        )


@statement(dialect=dialect)
class ConditionalBranch(ir.Statement):
    name = "cond_br"
    traits = frozenset({ir.IsTerminator()})

    cond: ir.SSAValue = info.argument(types.Bool)
    then_arguments: tuple[ir.SSAValue, ...]
    else_arguments: tuple[ir.SSAValue, ...]

    then_successor: ir.Block = info.block()
    else_successor: ir.Block = info.block()

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="keyword"):
            printer.print_name(self)

        printer.plain_print(" ")
        printer.print(self.cond)

        with printer.rich(style="keyword"):
            printer.plain_print(" goto ")

        printer.plain_print(printer.state.block_id[self.then_successor])
        printer.plain_print("(")
        printer.print_seq(self.then_arguments, delim=", ")
        printer.plain_print(")")

        with printer.rich(style="keyword"):
            printer.plain_print(" else ")

        printer.plain_print(printer.state.block_id[self.else_successor])
        printer.plain_print("(")
        printer.print_seq(self.else_arguments, delim=", ")
        printer.plain_print(")")

    def verify(self) -> None:
        return



from typing import cast

from kirin import rewrite
from kirin.prelude import basic
from kirin.dialects import py, func
from kirin.dialects.func.rewrite import closurefield


def test_rewrite_closure_inner_lambda():
    @basic
    def outer(y: int):
        def inner(x: int):
            return x * y + 1

        return inner

    inner_ker = outer(y=10)

    @basic
    def main_lambda(z: int):
        return inner_ker(z)

    main_invoke = main_lambda.code.regions[0].blocks[0].stmts.at(0)
    inner_lambda = cast(func.Invoke, main_invoke).callee.code
    inner_getfield_stmt = inner_lambda.regions[0].blocks[0].stmts.at(0)
    assert isinstance(
        inner_getfield_stmt, func.GetField
    ), "expected GetField before rewrite"

    rewrite.Walk(closurefield.ClosureField()).rewrite(main_lambda.code)

    inner_getfield_stmt = inner_lambda.regions[0].blocks[0].stmts.at(0)
    assert isinstance(
        inner_getfield_stmt, py.Constant
    ), "GetField should be lowered to Constant"


def test_rewrite_closure_no_fields():
    @basic
    def bar():
        def goo(x: int):
            a = (3, 4)
            return a[0]

        def boo(y):
            return goo(y) + 1

        return boo(4)

    before = bar.code.regions[0].blocks[0].stmts.at(0)
    rewrite.Walk(closurefield.ClosureField()).rewrite(bar.code)
    after = bar.code.regions[0].blocks[0].stmts.at(0)
    assert before is after



from dataclasses import field, dataclass

from kirin.passes import Pass
from kirin.rewrite import (
    Walk,
    Chain,
    Inline,
    Fixpoint,
    Call2Invoke,
    ConstantFold,
    CFGCompactify,
    InlineGetItem,
    InlineGetField,
    DeadCodeElimination,
)
from kirin.ir.method import Method
from kirin.rewrite.abc import RewriteResult
from kirin.passes.hint_const import HintConst


@dataclass
class Fold(Pass):
    hint_const: HintConst = field(init=False)

    def __post_init__(self):
        self.hint_const = HintConst(self.dialects)
        self.hint_const.no_raise = self.no_raise

    def unsafe_run(self, mt: Method) -> RewriteResult:
        result = self.hint_const.unsafe_run(mt)
        rule = Chain(
            ConstantFold(),
            Call2Invoke(),
            InlineGetField(),
            InlineGetItem(),
            DeadCodeElimination(),
        )
        result = Fixpoint(Walk(rule)).rewrite(mt.code).join(result)
        result = Walk(Inline(lambda _: True)).rewrite(mt.code).join(result)
        result = Fixpoint(CFGCompactify()).rewrite(mt.code).join(result)
        return result



from kirin.ir import DialectGroup
from kirin.prelude import basic
from kirin.analysis import const
from kirin.dialects import cf, func
from kirin.dialects.py import base


def test_union():
    group_a = DialectGroup([base, cf])
    group_b = DialectGroup([base, cf])
    group_c = DialectGroup([base, func])
    group_d = DialectGroup([base, func, cf])

    target_a = group_a.union(group_b)
    target_b = group_a.union(group_c)
    assert target_a.data == group_a.data
    assert target_b.data == group_d.data

    target_a_repr = repr(target_a)
    assert "DialectGroup(" in target_a_repr
    assert base.dialect.name in target_a_repr
    assert cf.dialect.name in target_a_repr

    target_b_repr = repr(target_b)
    assert "DialectGroup(" in target_b_repr
    assert base.dialect.name in target_b_repr
    assert cf.dialect.name in target_b_repr
    assert func.dialect.name in target_b_repr


def test_discard():
    group_a = DialectGroup([base, cf])
    group_c = DialectGroup([base, func])
    group_d = DialectGroup([base, func, cf])

    target_a = group_d.discard(cf)
    target_b = group_d.discard(func)
    assert target_a.data == group_c.data
    assert target_b.data == group_a.data

    target_a_repr = repr(target_a)
    assert "DialectGroup(" in target_a_repr
    assert base.dialect.name in target_a_repr
    assert func.dialect.name in target_a_repr

    target_b_repr = repr(target_b)
    assert "DialectGroup(" in target_b_repr
    assert base.dialect.name in target_b_repr
    assert cf.dialect.name in target_b_repr


def test_overwrite():
    @basic
    def foo(x):  # type: ignore
        return x * 2

    @basic
    def main(x):
        return x + foo(x)

    assert main(2) == 6

    @basic
    def foo(x):  # noqa: F811
        return x * 3

    assert main(2) == 8


def test_recompile():
    @basic
    def foo(x):  # type: ignore
        return x * 2

    @basic(fold=True)
    def main(x):
        return x + foo(x)

    ret = main.callable_region.blocks[0].stmts.at(0).results[0]
    assert isinstance(ret.hints.get("const"), const.Unknown)

    @basic
    def foo(x):  # noqa: F811
        return 3

    assert isinstance(ret.hints.get("const"), const.Value)
